import unittest
from shexer.shaper import Shaper
from test.const import G1, BASE_FILES, default_namespaces
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth
from shexer.consts import TURTLE

_BASE_DIR = BASE_FILES + "shape_map" + pth.sep

class TestRawShapeMap(unittest.TestCase):

    def test_node(self):
        shape_map = "<http://example.org/Jimmy>@<Person>"
        shaper = Shaper(graph_file_input=G1,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True,
                        shape_map_raw=shape_map
                        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "a_node.shex",
                                                      str_target=str_result))

    def test_prefixed_node(self):
        shape_map = "ex:Jimmy@<Person>"
        shaper = Shaper(graph_file_input=G1,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True,
                        shape_map_raw=shape_map
                        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "a_node.shex",
                                                      str_target=str_result))

    def test_focus(self):
        shape_map = "{FOCUS a foaf:Person}@<Person>"
        shaper = Shaper(graph_file_input=G1,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True,
                        shape_map_raw=shape_map
                        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "focus_nodes.shex",
                                                      str_target=str_result))

    def test_focus_wildcard(self):
        shape_map = "{FOCUS foaf:name _}@<WithName>"
        shaper = Shaper(graph_file_input=G1,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True,
                        shape_map_raw=shape_map
                        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "focus_and_wildcard.shex",
                                                      str_target=str_result))

    def test_sparql_selector(self):
        shape_map = "SPARQL \"select ?p where { ?p a foaf:Person }\"@<Person>"
        shaper = Shaper(graph_file_input=G1,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True,
                        shape_map_raw=shape_map
                        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "focus_nodes.shex",
                                                      str_target=str_result))

    def test_several_shapemap_items(self):
        shape_map = "{FOCUS a foaf:Person}@<Person>\n{FOCUS a foaf:Document}@<Document>"
        shaper = Shaper(graph_file_input=G1,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True,
                        shape_map_raw=shape_map
                        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "several_shm_items.shex",
                                                      str_target=str_result))



from shexer.consts import RDF_TYPE

class NodeSelector(object):

    def __init__(self, raw_selector, sgraph):
        self._raw_selector = raw_selector
        self._sgraph = sgraph

    @property
    def sgraph(self):
        return self._sgraph

    def get_target_nodes(self):
        """
        It returns a list of target URIs. It may require to execute some query against a given endpoint

        :return:
        """
        raise NotImplementedError()

    def yield_graph_of_target_nodes(self, depth=1, classes_at_last_level=True, instantiation_property=RDF_TYPE ):
        for a_triple in self._sgraph.yield_p_o_triples_of_target_nodes(target_nodes=self.get_target_nodes(),
                                                                       depth=depth,
                                                                       classes_at_last_level=classes_at_last_level,
                                                                       instantiation_property=instantiation_property,
                                                                       already_visited=None):
            yield a_triple



    @property
    def raw_selector(self):
        return self._raw_selector


########################################################


class NodeSelectorNoSparql(NodeSelector):

    def __init__(self, raw_selector, sgraph, target_node):
        super().__init__(raw_selector=raw_selector, sgraph=sgraph)
        self._target_nodes = [target_node]

    def get_target_nodes(self):
        return self._target_nodes


########################################################



class NodeSelectorSparql(NodeSelector):

    def __init__(self, raw_selector, sgraph, sparql_query_selector, id_variable_query):
        super().__init__(raw_selector=raw_selector, sgraph=sgraph)
        self._sparql_query_selector = sparql_query_selector
        self._id_variable_query = id_variable_query
        self._target_nodes = None

    @property
    def sparql_query_selector(self):
        return self._sparql_query_selector

    def get_target_nodes(self):
        if self._target_nodes is None:
            self._target_nodes = self._solve_target_nodes_at_endpoint()
        return self._target_nodes



    def _solve_target_nodes_at_endpoint(self):
        return self._sgraph.query_single_variable(str_query=self._sparql_query_selector,
                                                  variable_id=self._id_variable_query)






from shexer.model.IRI import IRI
from shexer.utils.uri import remove_corners

def create_IRI_from_string(an_str):
    return IRI(content=an_str)


def create_IRIs_from_string_list(str_list):
    result = []
    for an_str in str_list:
        result.append(create_IRI_from_string(remove_corners(a_uri=an_str,
                                                            raise_error_if_no_corners=False)))
    return result




import unittest
from shexer.shaper import Shaper
from test.const import BASE_FILES, default_namespaces, G1_ALL_CLASSES_NO_COMMENTS
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth
from shexer.consts import NT, TURTLE

_BASE_DIR = BASE_FILES + "general" + pth.sep

class TestUrlGraphFormat(unittest.TestCase):

    def test_ttl(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        url_graph_input="https://raw.githubusercontent.com/DaniFdezAlvarez/shexerp3/develop/test/t_files/t_graph_1.ttl",
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))


    def test_nt(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        url_graph_input="https://raw.githubusercontent.com/DaniFdezAlvarez/shexerp3/develop/test/t_files/t_graph_1.nt",
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=NT,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))





from wlighter import WLighter


def wikidata_annotation(raw_input, string_return, out_file, format, rdfs_comments):
    wlig = WLighter(raw_input=raw_input,
                    format=format,
                    languages=["en"],
                    generate_rdfs_comments=rdfs_comments,
                    mode_column_aligned=True)
    result = wlig.annotate_all(out_file=out_file, string_return=string_return)
    if string_return:
        return result



import json

from shexer.consts import RDF_TYPE, SHAPES_DEFAULT_NAMESPACE
from shexer.core.shexing.strategy.direct_shexing_strategy import DirectShexingStrategy
from shexer.core.shexing.strategy.direct_and_inverse_shexing_strategy import DirectAndInverseShexingStrategy
from shexer.utils.target_elements import determine_original_target_nodes_if_needed
from shexer.utils.log import log_msg
from shexer.consts import RATIO_INSTANCES


class ClassShexer(object):

    def __init__(self, class_counts_dict, class_profile_dict=None, class_profile_json_file=None,
                 remove_empty_shapes=True, original_target_classes=None, original_shape_map=None,
                 discard_useless_constraints_with_positive_closure=True, keep_less_specific=True,
                 all_compliant_mode=True, instantiation_property=RDF_TYPE, disable_or_statements=True,
                 disable_comments=False, namespaces_dict=None, tolerance_to_keep_similar_rules=0,
                 allow_opt_cardinality=True, disable_exact_cardinality=False,
                 shapes_namespace=SHAPES_DEFAULT_NAMESPACE, inverse_paths=False,
                 decimals=-1, instances_report_mode=RATIO_INSTANCES, detect_minimal_iri=False,
                 class_min_iris_dict=None, allow_redundant_or=False):
        self._class_counts_dict = class_counts_dict
        self._class_profile_dict = class_profile_dict if class_profile_dict is not None else self._load_class_profile_dict_from_file(
            class_profile_json_file)
        self._class_min_iris_dict = class_min_iris_dict

        self._shapes_list = []
        self._remove_empty_shapes = remove_empty_shapes
        self._all_compliant_mode = all_compliant_mode
        self._disable_or_statements = disable_or_statements
        self._instantiation_property_str = str(instantiation_property)
        self._disable_comments = disable_comments
        self._discard_useless_positive_closures = discard_useless_constraints_with_positive_closure
        self._namespaces_dict = namespaces_dict if namespaces_dict is not None else {}
        self._keep_less_specific = keep_less_specific
        self._tolerance = tolerance_to_keep_similar_rules
        self._allow_opt_cardinality = allow_opt_cardinality
        self._disable_exact_cardinality = disable_exact_cardinality
        self._shapes_namespace = shapes_namespace
        self._decimals = decimals
        self._instances_report_mode = instances_report_mode
        self._detect_minimal_iri = detect_minimal_iri
        self._allow_redundant_or = allow_redundant_or

        self._original_target_nodes = determine_original_target_nodes_if_needed(remove_empty_shapes=remove_empty_shapes,
                                                                                original_target_classes=original_target_classes,
                                                                                original_shape_map=original_shape_map,
                                                                                shapes_namespace=shapes_namespace)
        self._strategy = DirectShexingStrategy(self) if not inverse_paths \
            else DirectAndInverseShexingStrategy(self)

    def shex_classes(self, acceptance_threshold=0,
                     verbose=False):
        log_msg(verbose=verbose,
                msg="Starting shape extraction...")
        self._build_shapes(acceptance_threshold)
        log_msg(verbose=verbose,
                msg="Shape drafts built. Sorting constraints...")
        self._sort_shapes()
        log_msg(verbose=verbose,
                msg="Constraints sorted. Adjusting cardinalities...")
        self._set_valid_constraints_of_shapes()
        log_msg(verbose=verbose,
                msg="Cardinalities adjusted. Cleaning empty shapes if needed...")
        self._clean_empty_shapes()
        log_msg(verbose=verbose,
                msg="No more shapes to clean. {} definitive shapes".format(len(self._shapes_list)))
        return self._shapes_list

    def _set_valid_constraints_of_shapes(self):
        for a_shape in self._shapes_list:
            self._strategy.set_valid_shape_constraints(a_shape)

    def _build_shapes(self, acceptance_threshold):
        for a_shape in self._strategy.yield_base_shapes(acceptance_threshold=acceptance_threshold):
            self._shapes_list.append(a_shape)

    def _sort_shapes(self):
        for a_shape in self._shapes_list:
            a_shape.sort_statements(reverse=True,
                                    callback=self._value_to_compare_statements)

    def _clean_empty_shapes(self):
        if not self._remove_empty_shapes:
            return
        shapes_to_remove = self._detect_shapes_to_remove()

        while (len(shapes_to_remove) != 0):
            self._iteration_remove_empty_shapes(shapes_to_remove)
            shapes_to_remove = self._detect_shapes_to_remove()

    def _detect_shapes_to_remove(self):
        result = set()
        for a_shape in self._shapes_list:
            if a_shape.n_statements == 0:
                result.add(a_shape.name)
        return result

    def _iteration_remove_empty_shapes(self, shape_names_to_remove):
        self._remove_shapes_without_statements(shape_names_to_remove)
        self._remove_statements_to_gone_shapes(shape_names_to_remove)


    def _remove_statements_to_gone_shapes(self, shape_names_to_remove):
        for a_shape in self._shapes_list:
            self._strategy.remove_statements_to_gone_shapes(a_shape, shape_names_to_remove)

    def _remove_shapes_without_statements(self, shape_names_to_remove):
        new_shape_list = []
        for a_shape in self._shapes_list:
            if not a_shape.name in shape_names_to_remove:
                new_shape_list.append(a_shape)
        self._shapes_list = new_shape_list

    def _value_to_compare_statements(self, a_statement):
        return a_statement.probability

    @staticmethod
    def _load_class_profile_dict_from_file(source_file):
        with open(source_file, "r") as in_stream:
            return json.load(in_stream)





import unittest
from shexer.shaper import Shaper
from test.const import default_namespaces
from test.t_utils import number_of_shapes


class TestDisableEndpointCache(unittest.TestCase):

    """
    Note! These tests are similar to the ones found in TestUrlInput.
    We are only checking that if you try the same stuff without the cache, this will still work.


    These tests cannot be too precise, since they are made against "live" content in Wikidata and DBpedia.

    """

    def test_wikidata(self):
        shape_map_raw = "SPARQL'select ?p where " \
                        "{ ?p <http://www.wikidata.org/prop/direct/P31> <http://www.wikidata.org/entity/Q14660> } " \
                        "LIMIT 1'@<Flag>"
        shaper = Shaper(shape_map_raw=shape_map_raw,
                        url_endpoint="https://query.wikidata.org/sparql",
                        namespaces_dict=default_namespaces(),
                        instantiation_property="http://www.wikidata.org/prop/direct/P31",
                        disable_comments=True,
                        depth_for_building_subgraph=1,
                        track_classes_for_entities_at_last_depth_level=False,
                        all_classes_mode=False,
                        disable_endpoint_cache=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(number_of_shapes(str_result) == 1)

    def test_dbpedia(self):
        shape_map_raw = "SPARQL'select ?s where {?s a <http://dbpedia.org/ontology/Person>} LIMIT 1'@<Flag>"
        shaper = Shaper(shape_map_raw=shape_map_raw,
                        url_endpoint="https://dbpedia.org/sparql",
                        namespaces_dict=default_namespaces(),
                        disable_comments=True,
                        depth_for_building_subgraph=1,
                        track_classes_for_entities_at_last_depth_level=False,
                        all_classes_mode=False,
                        disable_endpoint_cache=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(number_of_shapes(str_result) == 1)

    def test_all_classes_mode(self):
        # shape_map_raw = "SPARQL'select ?s where {?s a <http://dbpedia.org/ontology/Person>} LIMIT 1'@<Flag>"
        shaper = Shaper(all_classes_mode=True,
                        url_endpoint="https://agrovoc.fao.org/sparql",
                        namespaces_dict=default_namespaces(),
                        disable_comments=True,
                        depth_for_building_subgraph=1,
                        track_classes_for_entities_at_last_depth_level=False,
                        limit_remote_instances=5,
                        disable_endpoint_cache=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(number_of_shapes(str_result) > 2)
        pass  #



from shexer.core.instances.annotators.base_annotator import BaseAnnotator
from shexer.core.instances.annotators.annotator_tracking_instances import AnnotatorTrackingInstances

def get_proper_annotator(track_hierarchies, instance_tracker_ref):
    return BaseAnnotator(instance_tracker_ref) if not track_hierarchies \
        else AnnotatorTrackingInstances(instance_tracker_ref)


import unittest
from shexer.shaper import Shaper
from test.const import default_namespaces
from test.t_utils import number_of_shapes, shape_contains_constraint


class TestDepthBuildingSubGraph(unittest.TestCase):

    """
    These test cannot be too precise, since they are made against "live" content in Wikidata.
    Also, two of them are commented, as they consume too much time to finish

    """

    def test_1_not_all_classes_format(self):
        shape_map_raw = "SPARQL'select ?p where " \
                        "{ ?p <http://www.wikidata.org/prop/direct/P31> <http://www.wikidata.org/entity/Q14660> } " \
                        "LIMIT 10'@<Flag>"
        shaper = Shaper(shape_map_raw=shape_map_raw,
                        url_endpoint="https://query.wikidata.org/sparql",
                        namespaces_dict=default_namespaces(),
                        instantiation_property="http://www.wikidata.org/prop/direct/P31",
                        disable_comments=True,
                        depth_for_building_subgraph=1,
                        track_classes_for_entities_at_last_depth_level=False,
                        all_classes_mode=False)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(number_of_shapes(str_result) == 1)
        self.assertTrue(shape_contains_constraint(target_str=str_result,
                                                  shape="<Flag>",
                                                  constraint="<http://www.wikidata.org/prop/direct/P31>  "
                                                             "[<http://www.wikidata.org/entity/Q14660>]"))

    def test_1_all_classes_format(self):
        shape_map_raw = "SPARQL'select ?p where " \
                        "{ ?p <http://www.wikidata.org/prop/direct/P31> <http://www.wikidata.org/entity/Q14660> } " \
                        "LIMIT 10'@<Flag>"
        shaper = Shaper(shape_map_raw=shape_map_raw,
                        url_endpoint="https://query.wikidata.org/sparql",
                        namespaces_dict=default_namespaces(),
                        instantiation_property="http://www.wikidata.org/prop/direct/P31",
                        disable_comments=True,
                        depth_for_building_subgraph=1,
                        track_classes_for_entities_at_last_depth_level=False,
                        all_classes_mode=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(number_of_shapes(str_result) > 1)
        self.assertTrue(number_of_shapes(str_result) < 10)
        self.assertTrue(shape_contains_constraint(target_str=str_result,
                                                  shape="<Flag>",
                                                  constraint="<http://www.wikidata.org/prop/direct/P31>  "
                                                             "[<http://www.wikidata.org/entity/Q14660>]"))

    # def test_2_not_all_classes_format(self):
    #     shape_map_raw = "SPARQL'select ?p where " \
    #                     "{ ?p <http://www.wikidata.org/prop/direct/P31> <http://www.wikidata.org/entity/Q14660> } " \
    #                     "LIMIT 10'@<Flag>"
    #     shaper = Shaper(shape_map_raw=shape_map_raw,
    #                     url_endpoint="https://query.wikidata.org/sparql",
    #                     namespaces_dict=default_namespaces(),
    #                     instantiation_property="http://www.wikidata.org/prop/direct/P31",
    #                     disable_comments=True,
    #                     depth_for_building_subgraph=2,
    #                     track_classes_for_entities_at_last_depth_level=False,
    #                     all_classes_mode=False)
    #     str_result = shaper.shex_graph(string_output=True)
    #     self.assertTrue(number_of_shapes(str_result) == 1)
    #     self.assertTrue(shape_contains_constraint(target_str=str_result,
    #                                               shape="<Flag>",
    #                                               constraint="<http://www.wikidata.org/prop/direct/P31>  "
    #                                                          "[<http://www.wikidata.org/entity/Q14660>]"))
    #
    # def test_2_all_classes_format(self):
    #     shape_map_raw = "SPARQL'select ?p where " \
    #                     "{ ?p <http://www.wikidata.org/prop/direct/P31> <http://www.wikidata.org/entity/Q14660> } " \
    #                     "LIMIT 10'@<Flag>"
    #     shaper = Shaper(shape_map_raw=shape_map_raw,
    #                     url_endpoint="https://query.wikidata.org/sparql",
    #                     namespaces_dict=default_namespaces(),
    #                     instantiation_property="http://www.wikidata.org/prop/direct/P31",
    #                     disable_comments=True,
    #                     depth_for_building_subgraph=2,
    #                     track_classes_for_entities_at_last_depth_level=False,
    #                     all_classes_mode=True)
    #     str_result = shaper.shex_graph(string_output=True)
    #     self.assertTrue(number_of_shapes(str_result) > 10)
    #     self.assertTrue(shape_contains_constraint(target_str=str_result,
    #                                               shape="<Flag>",
    #                                               constraint="<http://www.wikidata.org/prop/direct/P31>  "
    #                                                          "[<http://www.wikidata.org/entity/Q14660>]"))


from shexer.model.graph.endpoint_sgraph import EndpointSGraph

def get_remote_graph_if_needed(endpoint_url, store_locally):
    if endpoint_url is None:
        return None
    return EndpointSGraph(endpoint_url=endpoint_url,
                          store_locally=store_locally)


from shexer.core.profiling.strategy.abstract_feature_direction_strategy import AbstractFeatureDirectionStrategy
from shexer.core.profiling.consts import _S, _P, _O, POS_FEATURES_INVERSE, POS_CLASSES
from shexer.model.IRI import IRI_ELEM_TYPE

_C_MAP_POS_DIRECT = 0
_C_MAP_POS_INVERSE = 1


class IncludeReverseFeaturesStrategy(AbstractFeatureDirectionStrategy):



    def __init__(self, class_profiler):
        super().__init__(class_profiler)
        self._set_annotation_methods()


    def adapt_instances_dict(self):
        for a_subj_key in self._i_dict:
            self._i_dict[a_subj_key] = (self._i_dict[a_subj_key], {}, {})


    def is_a_relevant_triple(self, a_triple):
        target_elems = (a_triple[_S], a_triple[_O])
        for elem in target_elems:
            if self._is_relevant_instance(elem):
                return True
        return False


    def annotate_triple_features(self, a_triple):
        raise NotImplementedError()


    def init_annotated_targets(self):
        for an_instance, class_list in self._i_dict.items():
            for a_class in class_list:
                if a_class not in self._c_shapes_dict:
                    self._c_shapes_dict[a_class] = ({}, {})
                    self._c_counts[a_class] = 0
                self._c_counts[a_class] += 1

    def init_original_targets(self):
        if self._original_raw_target_classes:
            for a_class in self._original_raw_target_classes:
                self._c_shapes_dict[a_class] = ({}, {})
                self._c_counts[a_class] = 0

    def annotate_instance_features(self, an_instance):
        self._annotate_2d_direct_instance_features(an_instance)
        self._annotate_2d_inverse_instance_features(an_instance)

    def has_shape_annotated_features(self, shape_label):
        if shape_label not in self._c_shapes_dict:
            return False
        return len(self._c_shapes_dict[shape_label][_C_MAP_POS_DIRECT]) > 0 or \
               len(self._c_shapes_dict[shape_label][_C_MAP_POS_INVERSE]) > 0

    def _annotate_2d_direct_instance_features(self, an_instance):
        direct_feautres_3tuple = self._infer_direct_3tuple_features(an_instance)
        for a_class in self._i_dict[an_instance][POS_CLASSES]:
            self._annotate_2d_direct_instance_features_for_class(a_class, direct_feautres_3tuple)

    def _annotate_2d_inverse_instance_features(self, an_instance):
        inverse_feautres_3tuple = self._infer_inverse_3tuple_features(an_instance)
        for a_class in self._i_dict[an_instance][POS_CLASSES]:
            self._annotate_2d_inverse_instance_features_for_class(a_class, inverse_feautres_3tuple)

    def _infer_inverse_3tuple_features(self, an_instance):
        result = []
        for a_prop in self._i_dict[an_instance][POS_FEATURES_INVERSE]:
            for a_type in self._i_dict[an_instance][POS_FEATURES_INVERSE][a_prop]:
                for a_valid_cardinality in self._infer_valid_cardinalities(a_prop,
                                                                           self._i_dict[an_instance][POS_FEATURES_INVERSE][a_prop][a_type]):
                    result.append( (a_prop, a_type, a_valid_cardinality) )
        return result

    def _annotate_2d_direct_instance_features_for_class(self, a_class, features_3tuple):
        for a_feature_3tuple in features_3tuple:
            self._introduce_needed_direct_elements_in_2d_shape_classes_dict(a_class, a_feature_3tuple)
            # 3tuple: 0->str_prop, 1->str_type, 2->cardinality
            self._c_shapes_dict[a_class][_C_MAP_POS_DIRECT][a_feature_3tuple[0]][a_feature_3tuple[1]][a_feature_3tuple[2]] += 1

    def _annotate_2d_inverse_instance_features_for_class(self, a_class, features_3tuple):
        for a_feature_3tuple in features_3tuple:
            self._introduce_needed_inverse_elements_in_2d_shape_classes_dict(a_class, a_feature_3tuple)
            # 3tuple: 0->str_prop, 1->str_type, 2->cardinality
            self._c_shapes_dict[a_class][_C_MAP_POS_INVERSE][a_feature_3tuple[0]][a_feature_3tuple[1]][a_feature_3tuple[2]] += 1

    def _introduce_needed_direct_elements_in_2d_shape_classes_dict(self, a_class, a_feature_3tuple):
        str_prop = a_feature_3tuple[0]
        str_type = a_feature_3tuple[1]
        cardinality = a_feature_3tuple[2]
        if str_prop not in self._c_shapes_dict[a_class][_C_MAP_POS_DIRECT]:
            self._c_shapes_dict[a_class][_C_MAP_POS_DIRECT][str_prop] = {}
        if str_type not in self._c_shapes_dict[a_class][_C_MAP_POS_DIRECT][str_prop]:
            self._c_shapes_dict[a_class][_C_MAP_POS_DIRECT][str_prop][str_type] = {}
        if cardinality not in self._c_shapes_dict[a_class][_C_MAP_POS_DIRECT][str_prop][str_type]:
            self._c_shapes_dict[a_class][_C_MAP_POS_DIRECT][str_prop][str_type][cardinality] = 0

    def _introduce_needed_inverse_elements_in_2d_shape_classes_dict(self, a_class, a_feature_3tuple):
        str_prop = a_feature_3tuple[0]
        str_type = a_feature_3tuple[1]
        cardinality = a_feature_3tuple[2]
        if str_prop not in self._c_shapes_dict[a_class][_C_MAP_POS_INVERSE]:
            self._c_shapes_dict[a_class][_C_MAP_POS_INVERSE][str_prop] = {}
        if str_type not in self._c_shapes_dict[a_class][_C_MAP_POS_INVERSE][str_prop]:
            self._c_shapes_dict[a_class][_C_MAP_POS_INVERSE][str_prop][str_type] = {}
        if cardinality not in self._c_shapes_dict[a_class][_C_MAP_POS_INVERSE][str_prop][str_type]:
            self._c_shapes_dict[a_class][_C_MAP_POS_INVERSE][str_prop][str_type][cardinality] = 0


    def _annotate_target_object(self, a_triple):  # TODO: refactor here, place this in superclass and parametrize positions
        str_obj = a_triple[_O].iri
        str_prop = a_triple[_P].iri
        type_subj = self._decide_type_elem(a_triple[_S], str_prop)

        subj_shapes = [] if type_subj != IRI_ELEM_TYPE else self._decide_shapes_elem(a_triple[_S].iri)

        self._introduce_needed_elements_in_shape_instances_dict_for_obj(str_obj=str_obj,
                                                                        str_prop=str_prop,
                                                                        type_subj=type_subj,
                                                                        subj_shapes=subj_shapes)
        self._i_dict[str_obj][POS_FEATURES_INVERSE][str_prop][type_subj] += 1
        for a_shape in subj_shapes:
            self._i_dict[str_obj][POS_FEATURES_INVERSE][str_prop][a_shape] += 1

    def _introduce_needed_elements_in_shape_instances_dict_for_obj(self, str_obj, str_prop, type_subj, subj_shapes):
        if str_prop not in self._i_dict[str_obj][POS_FEATURES_INVERSE]:
            self._i_dict[str_obj][POS_FEATURES_INVERSE][str_prop] = {}
        if type_subj not in self._i_dict[str_obj][POS_FEATURES_INVERSE][str_prop]:
            self._i_dict[str_obj][POS_FEATURES_INVERSE][str_prop][type_subj] = 0
        for a_shape in subj_shapes:
            if a_shape not in self._i_dict[str_obj][POS_FEATURES_INVERSE][str_prop]:
                self._i_dict[str_obj][POS_FEATURES_INVERSE][str_prop][a_shape] = 0

    def _set_annotation_methods(self):
        if not self._examples_mode:
            self.annotate_triple_features = self._annotate_triple_features_no_examples
        else:
            self.annotate_triple_features = self._annotate_triple_features_with_examples

    def _annotate_triple_features_with_examples(self, a_triple):
        if self._is_relevant_instance(a_triple[_S]):
            self._annotate_target_subject(a_triple)
            self._annotate_example_subject_inverse_paths(a_triple)
        if self._is_relevant_instance(a_triple[_O]):
            self._annotate_target_object(a_triple)
            self._annotate_example_object_inverse_paths(a_triple)

    def _annotate_triple_features_no_examples(self, a_triple):
        if self._is_relevant_instance(a_triple[_S]):
            self._annotate_target_subject(a_triple)
        if self._is_relevant_instance(a_triple[_O]):
            self._annotate_target_object(a_triple)

    def _annotate_example_subject_inverse_paths(self, a_triple):
        for a_class_key in self._i_dict[str(a_triple[_S])][POS_CLASSES]:
            if not self._shape_feature_examples.has_constraint_example(shape_id=a_class_key,
                                                                       prop_id=str(a_triple[_P]),
                                                                       inverse=False):
                self._shape_feature_examples.set_constraint_example(shape_id=a_class_key,
                                                                    prop_id=str(a_triple[_P]),
                                                                    example=str(a_triple[_O]),
                                                                    inverse=False)

    def _annotate_example_object_inverse_paths(self, a_triple):
        for a_class_key in self._i_dict[str(a_triple[_O])][POS_CLASSES]:
            if not self._shape_feature_examples.has_constraint_example(shape_id=a_class_key,
                                                                       prop_id=str(a_triple[_P]),
                                                                       inverse=True):
                self._shape_feature_examples.set_constraint_example(shape_id=a_class_key,
                                                                    prop_id=str(a_triple[_P]),
                                                                    example=str(a_triple[_S]),
                                                                    inverse=True)





from shexer.core.profiling.class_profiler import ClassProfiler

from shexer.consts import SHAPES_DEFAULT_NAMESPACE
from shexer.core.profiling.consts import RDF_TYPE_STR
from shexer.utils.log import log_msg
from shexer.model.graph.endpoint_sgraph import EndpointSGraph
from shexer.utils.triple_yielders import tune_token, tune_prop
from shexer.core.instances.pconsts import _S, _P, _O

class FederatedSourceClassProfiler(ClassProfiler):

    def __init__(self, triples_yielder, instances_dict, instantiation_property_str=RDF_TYPE_STR,
                 remove_empty_shapes=True, original_target_classes=None, original_shape_map=None,
                 shapes_namespace=SHAPES_DEFAULT_NAMESPACE, inverse_paths=False, detect_minimal_iri=False,
                 examples_mode=None, list_of_federated_objects=None):
        super().__init__(triples_yielder, instances_dict, instantiation_property_str, remove_empty_shapes,
                         original_target_classes, original_shape_map, shapes_namespace, inverse_paths,
                         detect_minimal_iri, examples_mode)

        self._list_of_federated_objects = list_of_federated_objects


    # TODO CHECK ENTIRE PRIFILING PROCESS. maybe there is only need to redefine the yield_triples as it's already done.
    # def profile_classes(self, verbose):
        # log_msg(verbose=verbose,
        #         msg="Starting class profiler...")
        # self._init_class_counts_and_shape_dict()
        # log_msg(verbose=verbose,
        #         msg="Instance counts completed. Annotating instance features...")
        # self._adapt_instances_dict()
        # self._build_shape_of_instances()
        # log_msg(verbose=verbose,
        #         msg="Instance features annotated. Number of relevant triples computed: {}. "
        #             "Building shape profiles...".format(self._relevant_triples))
        #
        # self._build_class_profile()
        # log_msg(verbose=verbose,
        #         msg="Draft shape profiles built. Cleaning shape profiles...")
        # self._clean_class_profile()
        # log_msg(verbose=verbose,
        #         msg="Shape profiles done. Working with {} shapes.".format(len(self._classes_shape_dict)))
        # if self._detect_minimal_iri or self._examples_mode in [SHAPE_EXAMPLES, ALL_EXAMPLES]:
        #     log_msg(verbose=verbose,
        #             msg="Detecting example features for each shape...")
        #     self._init_anotation_example_method()
        #     self._detect_example_features()
        #     log_msg(verbose=verbose,
        #             msg="Mimimal IRIs detected...")
        # return self._classes_shape_dict, self._class_counts, \
        #     self._shape_feature_examples if (self._detect_minimal_iri or self._examples_mode is not None) else None




    # def _complete_class_counts_with_fed_sources(self):
    #     # There whould not be any new class, shape, as they are all mentioned in the base dict.
    #     # All we have to do here is to add class counts wlaking the fed_instance_dicts
    #     for a_fed_source in self._list_of_federated_objects:
    #         self._complete_class_count_of_fed_source(a_fed_source)
    #
    # def _complete_class_count_of_fed_source(self, federated_source_obj):
    #     for an_instance, class_list in federated_source_obj.instances_dict.items():
    #         for a_class in class_list:
    #             # if a_class not in self._c_shapes_dict:
    #             #     self._c_shapes_dict[a_class] = {}
    #             #     self._c_counts[a_class] = 0
    #             self._c_counts[a_class] += 1
    #
    # def _complete_instance_dict_adaptation(self):
    #     for a_fed_source in self._list_of_federated_objects:
    #         if not self._inverse_paths:
    #             self._adapt_i_dict_direct_of_a_


    def _yield_relevant_triples(self):
        for a_triple in self._triples_yielder.yield_triples():
            if self._strategy.is_a_relevant_triple(a_triple):
                yield a_triple

        for a_fed_source in self._list_of_federated_objects:
            for a_triple in self._yield_triples_of_fed_source(a_fed_source):
                yield a_triple

    def _yield_triples_of_fed_source(self, fed_source):
        end_graph = EndpointSGraph(fed_source.endpoint_url, store_locally=False)
        if not self._inverse_paths:
            triples = self._direct_target_triples_of_fed_source(end_graph, fed_source)
        else:
            triples = self._direct_and_inverse_target_triples_of_fed_source(end_graph, fed_source)
        for a_triple in triples:
            yield (
                tune_token(a_triple[_S]),
                tune_prop(a_triple[_P]),
                tune_token(a_triple[_O])
            )

    def _direct_target_triples_of_fed_source(self, end_graph, fed_source):
        triples = set()
        for a_target_node in fed_source.set_of_instances:
            for a_triple in end_graph.yield_p_o_triples_of_an_s(a_target_node):
                triples.add(a_triple)
        return triples

    def _direct_and_inverse_target_triples_of_fed_source(self, end_graph, fed_source):
        triples = set()
        for a_target_node in fed_source.set_of_instances:
            for a_triple in end_graph.yield_p_o_triples_of_an_s(a_target_node):
                triples.add(a_triple)
            for a_triple in end_graph.yield_s_p_triples_of_an_o(a_target_node):
                triples.add(a_triple)
        return triples



import unittest
from shexer.shaper import Shaper
from test.const import BASE_FILES, default_namespaces
from test.t_utils import file_vs_str_exact_comparison
import os.path as pth
from shexer.consts import TURTLE

_BASE_DIR = BASE_FILES + "disable_comments" + pth.sep

class TestGraphFileInput(unittest.TestCase):

    def test_disable(self):
        shaper = Shaper(all_classes_mode=True,
                        graph_file_input=_BASE_DIR + "g2.ttl",
                        namespaces_dict=default_namespaces(),
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_exact_comparison(target_str=str_result,
                                                     file_path=_BASE_DIR + "g2_disable.shex"))

    def test_enable(self):
        shaper = Shaper(all_classes_mode=True,
                        graph_file_input=_BASE_DIR + "g2.ttl",
                        namespaces_dict=default_namespaces(),
                        input_format=TURTLE,
                        disable_comments=False)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_exact_comparison(target_str=str_result,
                                                     file_path=_BASE_DIR + "g2_enable.shex"))



import string
import random

_PRIORITY_PREFIXES_FOR_SHAPES = ["", "weso-s", "shapes", "w-shapes"]
_RAND_PREFIX_LENGHT = 3

def find_adequate_prefix_for_shapes_namespaces(current_namespace_prefix_dict, avoid_empty=False):
    curr_prefixes = current_namespace_prefix_dict.values()
    for a_prefix in _PRIORITY_PREFIXES_FOR_SHAPES[1 if avoid_empty else 0:]:
        if a_prefix not in curr_prefixes:
            return a_prefix
    # At this point, all the deff prefixes are used. So we generate a random one
    candidate = get_random_string(3)
    while candidate in curr_prefixes:
        candidate = get_random_string(3)
    return candidate

def get_random_string(length):
    result_str = ''.join(random.choice(string.ascii_lowercase) for i in range(length))
    return result_str



from shexer.core.shexing.strategy.abstract_shexing_strategy import AbstractShexingStrategy
from shexer.utils.shapes import build_shapes_name_for_class_uri
from shexer.model.statement import Statement
from shexer.model.shape import Shape

class DirectShexingStrategy(AbstractShexingStrategy):

    def __init__(self, class_shexer):
        super().__init__(class_shexer)
        self._class_profile_dict = self._class_shexer._class_profile_dict
        self._shapes_namespace = self._class_shexer._shapes_namespace
        self._class_counts_dict = self._class_shexer._class_counts_dict

    def remove_statements_to_gone_shapes(self, shape, shape_names_to_remove):
        shape.direct_statements = self._statements_without_shapes_to_remove(original_statements=shape.direct_statements,
                                                                            shape_names_to_remove=shape_names_to_remove)

    def set_valid_shape_constraints(self, shape):
        valid_statements = self._select_valid_statements_of_shape(shape.direct_statements)
        self._tune_list_of_valid_statements(valid_statements=valid_statements)
        shape.statements = valid_statements


    def _yield_base_shapes_direction_aware(self, acceptance_threshold):
        for a_class_key in self._class_profile_dict:
            name = build_shapes_name_for_class_uri(class_uri=a_class_key,
                                                   shapes_namespace=self._shapes_namespace)
            number_of_instances = float(self._class_counts_dict[a_class_key])
            statements = []
            for a_prop_key in self._class_profile_dict[a_class_key]:
                for a_type_key in self._class_profile_dict[a_class_key][a_prop_key]:
                    for a_cardinality in self._class_profile_dict[a_class_key][a_prop_key][a_type_key]:
                        n_occurences = self._class_profile_dict[a_class_key][a_prop_key][a_type_key][a_cardinality]
                        frequency = self._compute_frequency(number_of_instances,
                                                            n_occurences)
                        if frequency >= acceptance_threshold:
                            statements.append(Statement(st_property=a_prop_key,
                                                        st_type=a_type_key,
                                                        cardinality=a_cardinality,
                                                        probability=frequency,
                                                        n_occurences=n_occurences))

            yield Shape(name=name,
                        class_uri=a_class_key,
                        statements=statements,
                        n_instances=int(number_of_instances))




from rdflib import Graph, URIRef, Literal
from rdflib.compare import to_isomorphic, graph_diff
import re
from os.path import isfile, exists as file_exists
from os import remove

_BLANKS = re.compile("[ \t]+")
_LINE_JUMPS = re.compile("\n+")

_PREFIX = "PREFIX"
_BEG_SHAPE = "{"
_END_SHAPE = "}"
_BEG_OR = "("
_END_OR = ")"
_OR = " OR "


def get_namespaces_and_shapes_from_str(str_target, or_shapes=False):
    if or_shapes:
        return get_namespaces_and_shapes_from_str_with_or(str_target)
    namespaces = []
    shapes = {}
    last_line = ""
    current_shape = None

    for a_line in str_target.split("\n"):
        if a_line.startswith(_PREFIX):
            namespaces.append(a_line)
        elif a_line.startswith(_BEG_SHAPE):
            current_shape = last_line
            shapes[last_line] = []
        elif a_line.startswith(_END_SHAPE):
            current_shape = None
        elif current_shape is not None:
            shapes[current_shape].append(
                a_line.replace(";", "").strip())  # Avoid trailing ";", that can be there or not

        last_line = a_line  # Always execute

    return namespaces, shapes


def get_namespaces_and_shapes_from_str_with_or(str_target):
    namespaces = []
    shapes = {}
    last_line = ""
    current_shape = None

    or_mode = False
    current_or = None

    for a_line in str_target.split("\n"):
        a_line = a_line.strip()
        if a_line.startswith(_PREFIX):
            namespaces.append(a_line)
        elif a_line.startswith(_BEG_SHAPE):
            current_shape = last_line
            shapes[last_line] = []
        elif a_line.startswith(_END_SHAPE):
            current_shape = None
        elif a_line.startswith(_BEG_OR):
            or_mode = True
            current_or = ""
        elif or_mode:
            if not a_line.startswith(_END_OR):
                current_or = current_or + a_line
            else:
                or_mode = False
                shapes[current_shape].append(current_or)

        elif current_shape is not None:
            shapes[current_shape].append(
                a_line.replace(";", "").strip())  # Avoid trailing ";", that can be there or not

        last_line = a_line  # Always execute

    return namespaces, shapes


def unordered_lists_match(list1, list2):
    return set(list1) == set(list2)


def ordered_lists_match(list1, list2):
    return list1 == list2


def unordered_sets_match(sets1, sets2):
    for a_set1 in sets1:
        if a_set1 not in sets2:
            return False
    return True


def ordered_sets_match(sets1, sets2):
    if len(sets1) != len(sets2):
        return False
    for i in range(len(sets1)):
        if sets1[i] != sets2[i]:
            return False
    return True


def simple_and_or_str_constraints(
        str_constraints):  # TODO Fix here. Receive a dict, not a list. check call to this method
    simple_c = []
    or_c = []
    for a_c in str_constraints:
        if _OR in a_c:
            or_c.append(a_c)
        else:
            simple_c.append(a_c)
    return simple_c, or_c


def ordered_or_constraints_match(or_list_1, or_list_2):
    if len(or_list_1) != len(or_list_2):
        return False
    for i in range(len(or_list_1)):
        or_list_1[i] = set(or_list_1[i].split(_OR))
        or_list_2[i] = set(or_list_2[i].split(_OR))
    return ordered_sets_match(or_list_1, or_list_2)


def unordered_or_constraints_match(or_list_1, or_list_2):
    if len(or_list_1) != len(or_list_2):
        return False

    for i in range(len(or_list_1)):
        or_list_1[i] = set(or_list_1[i].split("  "))
        or_list_2[i] = set(or_list_2[i].split("  "))
    return unordered_sets_match(or_list_1, or_list_2)


def or_shapes_comparison(shapes1, shapes2, check_order):
    for a_key_label in shapes1:
        if a_key_label not in shapes2:
            return False
        simple_constraints1, or_constraints1 = simple_and_or_str_constraints(shapes1[a_key_label])
        simple_constraints2, or_constraints2 = simple_and_or_str_constraints(shapes2[a_key_label])
        if not check_order:
            if not unordered_lists_match(simple_constraints1, simple_constraints2):
                return False
            if not unordered_or_constraints_match(or_constraints1, or_constraints2):
                return False
        else:
            if not ordered_lists_match(simple_constraints1, simple_constraints2):
                return False
            if not ordered_or_constraints_match(or_constraints1, or_constraints2):
                return False
    return True


def namespaces_match(names1, names2):
    return unordered_lists_match(names1, names2)


def unsorted_constraints_comparison(shapes1, shapes2):
    for a_key_label in shapes1:
        if a_key_label not in shapes2:
            return False
        if not unordered_lists_match(shapes1[a_key_label], shapes2[a_key_label]):
            return False
    return True


def sorted_constraints_comparison(shapes1, shapes2):
    for a_key_label in shapes1:
        if a_key_label not in shapes2:
            return False
        if not ordered_lists_match(shapes1[a_key_label], shapes2[a_key_label]):
            return False
    return True


def simple_constraints_comparison(shapes1, shapes2, check_order=False):
    if not check_order:
        return unsorted_constraints_comparison(shapes1, shapes2)
    else:
        return sorted_constraints_comparison(shapes1, shapes2)


def shapes_match(shapes1, shapes2, or_shapes=False, check_order=False):
    if len(shapes1) != len(shapes2):
        return False
    if or_shapes:
        return or_shapes_comparison(shapes1, shapes2, check_order)
    return simple_constraints_comparison(shapes1, shapes2, check_order)


def complex_shape_comparison(str1, str2, or_shapes=False, check_order=False):
    namespaces1, shapes1 = get_namespaces_and_shapes_from_str(str1, or_shapes)
    namespaces2, shapes2 = get_namespaces_and_shapes_from_str(str2, or_shapes)

    if not namespaces_match(namespaces1, namespaces2):
        return False
    return shapes_match(shapes1, shapes2, or_shapes, check_order)


def normalize_str(str_target):
    result = str_target.strip()
    result = _BLANKS.sub(result, " ")
    return _LINE_JUMPS.sub(result, "\n")


def tunned_str_comparison(str1, str2, or_shapes=False, check_order=False):
    if normalize_str(str1) == normalize_str(str2):
        return True
    else:
        return complex_shape_comparison(str1, str2, or_shapes, check_order)


def file_vs_str_tunned_comparison(file_path, str_target, or_shapes=False, check_order=False):
    with open(file_path, "r") as in_stream:
        content = in_stream.read()
    return tunned_str_comparison(content, str_target, or_shapes, check_order)


def file_vs_str_exact_comparison(file_path, target_str):
    with open(file_path, "r") as in_stream:
        return in_stream.read().strip() == target_str.strip()


def filter_prefixes_str_shex(target_str):
    result = []
    lines = target_str.split("\n")
    counter = 0
    while len(lines) > counter:
        if not lines[counter].startswith("PREFIX"):
            break
        counter += 1
    return "\n".join(lines[counter:]).strip() if counter > 0 else target_str


def file_vs_str_shex_exact_comparison_excluding_prefixes(file_path, str_target):
    with open(file_path, "r") as in_stream:
        target_file_content = filter_prefixes_str_shex(in_stream.read().strip())
        target_str_content = filter_prefixes_str_shex(str_target.strip())

        return target_file_content == target_str_content


def file_vs_file_tunned_comparison(file_path1, file_path2, or_shapes=False):
    with open(file_path1, "r") as in_stream:
        content1 = in_stream.read()
    with open(file_path2, "r") as in_stream:
        content2 = in_stream.read()
    return tunned_str_comparison(content1, content2, or_shapes)


def number_of_shapes(target_str):
    counter = 0
    for a_line in target_str.split("\n"):
        if a_line.startswith(_BEG_SHAPE):
            counter += 1
    return counter


def shape_contains_constraint(target_str, shape, constraint):
    constraint = constraint.replace(";", "").strip()
    lines = target_str.split("\n")
    seeking_mode = False
    for i in range(len(lines)):
        if seeking_mode:
            if lines[i].replace(";", "").strip() == constraint:
                return True
            if lines[i].startswith(_END_SHAPE):
                return False
        if lines[i].startswith(_BEG_SHAPE) and shape == lines[i - 1].strip():
            seeking_mode = True
    return False


def graph_comparison_rdflib(g1, g2):
    iso1 = to_isomorphic(g1)
    iso2 = to_isomorphic(g2)
    both, in1, in2 = graph_diff(iso1, iso2)
    return len(both) == len(g1) and len(in1) == 0 and len(in2) == 0


def _tune_elem_for_rdflib_comparison(original_elem):
    if original_elem is None:
        return None
    elif original_elem.startswith("<"):
        return URIRef(original_elem[1:-1])
    else:  # It should be a literal
        return Literal(original_elem)


def rdflib_exist_triple(graph, t_subject, t_predicate, t_object):
    return any(
        [a_triple for a_triple in graph.triples(
            (_tune_elem_for_rdflib_comparison(t_subject),
             _tune_elem_for_rdflib_comparison(t_predicate),
             _tune_elem_for_rdflib_comparison(t_object)
             )
        )]
    )


def rdflib_counts_triples(graph, t_subject, t_predicate, t_object):
    return len(
        [a_triple for a_triple in graph.triples(
            (_tune_elem_for_rdflib_comparison(t_subject),
             _tune_elem_for_rdflib_comparison(t_predicate),
             _tune_elem_for_rdflib_comparison(t_object)
             )
        )]
    )


def graph_comparison_file_vs_str(file_path, str_target, format="turtle"):
    g1 = Graph()
    g1.parse(data=str_target, format=format)

    g2 = Graph()
    g2.parse(source=file_path, format=format)

    return graph_comparison_rdflib(g1, g2)


def text_contains_lines(text, list_lines):
    text = _BLANKS.sub(" ", text)
    for a_line in list_lines:
        a_line = _BLANKS.sub(" ", a_line)
        if a_line not in text:
            return False
        return True


def no_sharp_in_shepe_names(str_shapes):
    _, shapes = get_namespaces_and_shapes_from_str(str_shapes, or_shapes=False)
    for a_key_label in shapes:
        if '#' in a_key_label:
            return False
    return True


def check_file_exist(file_path):
    if not isfile(file_path):
        raise FileExistsError(f"The expected file was not found in disk: {file_path}")


def delete_file(file_path):
    if file_exists(file_path):
        try:
            remove(file_path)
        except:
            pass



from shexer.core.shexing.class_shexer import ClassShexer
from shexer.consts import RDF_TYPE, SHAPES_DEFAULT_NAMESPACE
from shexer.consts import RATIO_INSTANCES
from shexer.core.instances.pconsts import FEDERATION_TAG_MARK

_COMMENT_FED_PROPERTY = "# Constraint only observed in {}"

class ClassShexerFedSources(ClassShexer):


    def __init__(self, class_counts_dict, class_profile_dict=None, class_profile_json_file=None,
                 remove_empty_shapes=True, original_target_classes=None, original_shape_map=None,
                 discard_useless_constraints_with_positive_closure=True, keep_less_specific=True,
                 all_compliant_mode=True, instantiation_property=RDF_TYPE, disable_or_statements=True,
                 disable_comments=False, namespaces_dict=None, tolerance_to_keep_similar_rules=0,
                 allow_opt_cardinality=True, disable_exact_cardinality=False, shapes_namespace=SHAPES_DEFAULT_NAMESPACE,
                 inverse_paths=False, decimals=-1, instances_report_mode=RATIO_INSTANCES, detect_minimal_iri=False,
                 class_min_iris_dict=None, allow_redundant_or=False, fed_sources=None):
        super().__init__(class_counts_dict, class_profile_dict, class_profile_json_file, remove_empty_shapes,
                         original_target_classes, original_shape_map, discard_useless_constraints_with_positive_closure,
                         keep_less_specific, all_compliant_mode, instantiation_property, disable_or_statements,
                         disable_comments, namespaces_dict, tolerance_to_keep_similar_rules, allow_opt_cardinality,
                         disable_exact_cardinality, shapes_namespace, inverse_paths, decimals, instances_report_mode,
                         detect_minimal_iri, class_min_iris_dict, allow_redundant_or)
        self._fed_sources = fed_sources

    def shex_classes(self, acceptance_threshold=0,
                     verbose=False):
        result = super().shex_classes(acceptance_threshold=acceptance_threshold,
                                      verbose=verbose)
        self._annotate_properties_in_fed_shapes(result)
        return result


    def _annotate_properties_in_fed_shapes(self, list_of_shapes):
        pairs = self._find_pairs_of_origin_and_fed_shape(list_of_shapes)
        for a_origin_fed_pair in pairs:
            self._annotate_fed_properties_of_a_shape(origin_shape=a_origin_fed_pair[0],
                                                     fed_shape=a_origin_fed_pair[1])


    def _annotate_fed_properties_of_a_shape(self, origin_shape, fed_shape):
        fed_name_for_comments = fed_shape.class_uri[
                                fed_shape.class_uri.find(FEDERATION_TAG_MARK)+len(FEDERATION_TAG_MARK):]
        for a_fed_statement in fed_shape.direct_statements:
            only_in_fed = True
            for a_orig_statement in origin_shape.direct_statements:
                if a_fed_statement.st_property == self._instantiation_property_str:
                    if a_fed_statement.st_property == a_orig_statement.st_property and a_fed_statement.st_type == a_orig_statement.st_type:
                        only_in_fed = False
                        break
                else:
                    if a_fed_statement.st_property == a_orig_statement.st_property:
                        only_in_fed = False
                        break
            if only_in_fed:
                a_fed_statement.add_comment(_COMMENT_FED_PROPERTY.format(fed_name_for_comments))
        # TODO: same thing with inverse. Probably refactor
        for a_fed_statement in fed_shape.inverse_statements:
            only_in_fed = False
            for a_orig_statement in origin_shape.inverse_statements:
                if a_fed_statement.st_property == self._instantiation_property:
                    if a_fed_statement.st_property == a_orig_statement.st_property and a_fed_statement.st_type == a_orig_statement.st_type:
                        only_in_fed = True
                        break
                else:
                    if a_fed_statement.st_property == a_orig_statement.st_property:
                        only_in_fed = True
                        break
            if only_in_fed:
                a_fed_statement.add_comment(comment=_COMMENT_FED_PROPERTY.format(fed_name_for_comments),
                                            insert_first=False)

    def _find_pairs_of_origin_and_fed_shape(self, list_of_shapes):
        result = []
        for i in range(len(list_of_shapes)):
            a_shape = list_of_shapes[i]
            for a_potential_fed_shape in list_of_shapes[i+1:]:
                if a_potential_fed_shape.class_uri.startswith(a_shape.class_uri) and \
                        FEDERATION_TAG_MARK in a_potential_fed_shape.class_uri:
                    result.append((a_shape, a_potential_fed_shape))
        return result





from shexer.core.shexing.class_shexer import ClassShexer
from shexer.core.shexing.class_shexer_fed_sources import ClassShexerFedSources
from shexer.consts import RDF_TYPE, SHAPES_DEFAULT_NAMESPACE, RATIO_INSTANCES


def get_class_shexer(class_counts,
                     class_profile_dict,
                     remove_empty_shapes=True,
                     original_target_classes=None,
                     original_shape_map=None,
                     discard_useless_constraints_with_positive_closure=True,
                     keep_less_specific=True,
                     all_compliant_mode=True,
                     instantiation_property=RDF_TYPE,
                     disable_or_statements=True,
                     disable_comments=False,
                     namespaces_dict=None,
                     allow_opt_cardinality=True,
                     disable_exact_cardinality=False,
                     shapes_namespace=SHAPES_DEFAULT_NAMESPACE,
                     inverse_paths=False,
                     decimals=-1,
                     instances_report_mode=RATIO_INSTANCES,
                     detect_minimal_iri=False,
                     class_min_iris=None,
                     allow_redundant_or=False,
                     federated_sources=None):

    if federated_sources is None:
        return ClassShexer(
            class_counts_dict=class_counts,
            class_profile_dict=class_profile_dict,
            class_profile_json_file=None,
            remove_empty_shapes=remove_empty_shapes,
            original_target_classes=original_target_classes,
            original_shape_map=original_shape_map,
            discard_useless_constraints_with_positive_closure=discard_useless_constraints_with_positive_closure,
            keep_less_specific=keep_less_specific,
            all_compliant_mode=all_compliant_mode,
            instantiation_property=instantiation_property,
            disable_or_statements=disable_or_statements,
            disable_comments=disable_comments,
            namespaces_dict=namespaces_dict,
            allow_opt_cardinality=allow_opt_cardinality,
            disable_exact_cardinality=disable_exact_cardinality,
            shapes_namespace=shapes_namespace,
            inverse_paths=inverse_paths,
            instances_report_mode=instances_report_mode,
            decimals=decimals,
            detect_minimal_iri=detect_minimal_iri,
            class_min_iris_dict=class_min_iris,
            allow_redundant_or=allow_redundant_or
        )
    else:
        return ClassShexerFedSources(
            class_counts_dict=class_counts,
            class_profile_dict=class_profile_dict,
            class_profile_json_file=None,
            remove_empty_shapes=remove_empty_shapes,
            original_target_classes=original_target_classes,
            original_shape_map=original_shape_map,
            discard_useless_constraints_with_positive_closure=discard_useless_constraints_with_positive_closure,
            keep_less_specific=keep_less_specific,
            all_compliant_mode=all_compliant_mode,
            instantiation_property=instantiation_property,
            disable_or_statements=disable_or_statements,
            disable_comments=disable_comments,
            namespaces_dict=namespaces_dict,
            allow_opt_cardinality=allow_opt_cardinality,
            disable_exact_cardinality=disable_exact_cardinality,
            shapes_namespace=shapes_namespace,
            inverse_paths=inverse_paths,
            instances_report_mode=instances_report_mode,
            decimals=decimals,
            detect_minimal_iri=detect_minimal_iri,
            class_min_iris_dict=class_min_iris,
            allow_redundant_or=allow_redundant_or,
            fed_sources=federated_sources
        )



import unittest
from shexer.shaper import Shaper
from test.const import BASE_FILES, default_namespaces
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth
from shexer.consts import TURTLE

_BASE_DIR = BASE_FILES + "keep_less_specific" + pth.sep
_G1_SEV_NAMES = _BASE_DIR + "g1_several_names.ttl"

class TestKeepLessSpecific(unittest.TestCase):

    def test_keep_less_no_compliant(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=_G1_SEV_NAMES,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True,
                        all_instances_are_compliant_mode=False)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "keep_less_no_compliant.shex",
                                                      str_target=str_result))

    def test_no_keep_less_no_compliant(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=_G1_SEV_NAMES,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True,
                        all_instances_are_compliant_mode=False,
                        keep_less_specific=False)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "no_keep_less_no_compliant.shex",
                                                      str_target=str_result))

    def test_keep_less_compliant(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=_G1_SEV_NAMES,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True,
                        all_instances_are_compliant_mode=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "keep_less_compliant.shex",
                                                      str_target=str_result))

    def test_no_keep_less_compliant(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=_G1_SEV_NAMES,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True,
                        all_instances_are_compliant_mode=True,
                        keep_less_specific=False)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "no_keep_less_compliant.shex",
                                                      str_target=str_result))




from shexer.model.property import Property
from shexer.utils.uri import remove_corners
from shexer.utils.factories.h_tree import get_basic_h_tree

_TRACKERS_DISAM_COUNT = 0

_RDF_TYPE = Property(content="http://www.w3.org/1999/02/22-rdf-syntax-ns#type")
_RDFS_SUBCLASS_OF = Property(content="http://www.w3.org/2000/01/rdf-schema#subClassOf")

class AbstractInstanceTracker(object):

    def track_instances(self, verbose=False):
        raise NotImplementedError()


    @property
    def disambiguator_prefix(self):
        """
        It returns a str that may help for disambiguation purposes if the instance_tracker is used to produce dicts
        that may be integrated with other instance dicts and there should be any key colission.
        :return:
        """
        global _TRACKERS_DISAM_COUNT
        _TRACKERS_DISAM_COUNT += 1
        return self._specific_disambiguator_prefix() + str(_TRACKERS_DISAM_COUNT )

    def _specific_disambiguator_prefix(self):
        raise NotImplementedError()

    @staticmethod
    def _build_instances_dict():
        return {}  # Empty in every case. Instances, on the fly, will be the keys

    @staticmethod
    def _decide_instantiation_property(instantiation_property):
        if instantiation_property == None:
            return _RDF_TYPE
        if type(instantiation_property) == type(_RDF_TYPE):
            return instantiation_property
        if type(instantiation_property) == str:
            return Property(remove_corners(a_uri=instantiation_property,
                                           raise_error_if_no_corners=False))
        raise ValueError("Unrecognized param type to define instantiation property")

    def _reset_count(self):
        self._relevant_triples = 0
        self._not_relevant_triples = 0
        self._htree = get_basic_h_tree()




from shexer.core.instances.abstract_instance_tracker import AbstractInstanceTracker
from shexer.core.instances.pconsts import _S, _P, _O, FEDERATION_TAG_MARK
from shexer.io.sparql.query import query_endpoint_single_variable

_VARIABLE_NAME_QUERYING_REMOTE_SYNONYMS = "var"
_QUERY_SYNONYMS_ORIGIN_SUBJECT = "select ?s where {{ <{origin_node}> <{synonym_prop}> ?var .  }}"
_QUERY_SYNONYMS_ORIGIN_OBJECT = "select ?s where {{ ?var  <{synonym_prop}>  <{origin_node}> .  }}"

class FederatedSourceInstanceTracker(AbstractInstanceTracker):

    def __init__(self, instance_tracker, federated_source_objs):
        self._instance_tracker = instance_tracker
        self._federated_source_objs = federated_source_objs
        self._instances_dict_in_origin = None
        self._origin_triple_yielder = None
        self._query_to_find_synonyms = None  # may be changed later


    def track_instances(self, verbose=False):
        # TODO We have to do here a loop, walking through, potentially, several instances of federated sources.
        # Then, integrate all dicts into the origin_dict one.
        self._instances_dict_in_origin = self._instance_tracker.track_instances()
        self._origin_triple_yielder = self._instance_tracker._triples_yielder  # YES, let it be
        fed_dicts = []
        for a_fed_source in self._federated_source_objs:
            instances_dict_federated = self._build_fed_instances_dict(a_fed_source)
            # self._integrate_dicts(instances_dict_federated)
            a_fed_source.set_of_instances = set(instances_dict_federated.keys())
            fed_dicts.append(instances_dict_federated)
        for a_fed_dict in fed_dicts:
            self._integrate_dicts(a_fed_dict)
        return self._instances_dict_in_origin

    def _integrate_dicts(self, fed_source_dict):
        for a_key in fed_source_dict:
            self._instances_dict_in_origin[a_key] = fed_source_dict[a_key]

    def _build_fed_instances_dict(self, a_fed_source):
        fed_source_dict = {}
        for an_instance, a_synonym in self._find_synonyms(a_fed_source=a_fed_source, fed_source_dict=fed_source_dict):
            self._add_synonym_to_dicts(origin_instance=an_instance,
                                       synonym=a_synonym,
                                       fed_source_dict=fed_source_dict,
                                       a_fed_source=a_fed_source)
        return fed_source_dict

    def _add_synonym_to_dicts(self, origin_instance, synonym, fed_source_dict, a_fed_source):
        shape_labels = [self._adapted_shape_label(original_class=a_class,
                                                  fed_source=a_fed_source) for a_class in self._instances_dict_in_origin[origin_instance] if FEDERATION_TAG_MARK not in a_class]
        if synonym not in fed_source_dict:
            fed_source_dict[synonym] = []
        fed_source_dict[synonym].extend(shape_labels)
        self._instances_dict_in_origin[origin_instance].extend(shape_labels)

    def _adapted_shape_label(self, original_class, fed_source):
        return original_class + FEDERATION_TAG_MARK + fed_source.alias


    def _find_synonyms(self, a_fed_source, fed_source_dict):
        if not a_fed_source.link_in_federated_source:
            for an_instance_synonym_pair in self._find_synonyms_in_origin(a_fed_source):
                yield an_instance_synonym_pair
        else:
            for an_instance_synonym_pair in self._find_synonyms_in_fed_source(a_fed_source, fed_source_dict):
                yield an_instance_synonym_pair

    def _find_synonyms_in_fed_source(self, a_fed_source):
        self._query_to_find_synonyms = _QUERY_SYNONYMS_ORIGIN_SUBJECT \
            if a_fed_source.origin_position_in_triple == _S \
            else _QUERY_SYNONYMS_ORIGIN_OBJECT
        keys = self._instances_dict_in_origin.keys()
        for an_instance in keys:
            for a_synonym in self._query_remote_synonyms(target_instance=an_instance,
                                                         fed_source=a_fed_source):
                yield an_instance, a_synonym

    def _query_remote_synonyms(self, target_instance, fed_source):
        return query_endpoint_single_variable(endpoint_url=fed_source.endpoint_url,
                                              str_query=self._query_to_find_synonyms.format(origin_node=target_instance,
                                                                                            synonym_prop=fed_source.property_link),
                                              variable_id=_VARIABLE_NAME_QUERYING_REMOTE_SYNONYMS)

    def _find_synonyms_in_origin(self, a_fed_source):
        """
        It yields 2-tuples where where:
        - 0, instance (origin source)
        - 1, synonym (federated source)
        """
        instance_position = a_fed_source.origin_position_in_triple
        synonym_position = _S if a_fed_source.origin_position_in_triple == _O else _O
        for a_triple in self._origin_triple_yielder.yield_triples():
            if a_triple[_P].iri == a_fed_source.property_link:
                if a_triple[instance_position].iri in self._instances_dict_in_origin:
                    yield a_triple[instance_position].iri, a_triple[synonym_position].iri




from shexer.utils.log import log_msg
from shexer.utils.triple_yielders import tune_token, tune_prop
from shexer.io.graph.yielder.base_triples_yielder import BaseTriplesYielder


class TsvNtTriplesYielder(BaseTriplesYielder):

    def __init__(self, source_file, allow_untyped_numbers=False, raw_graph=None,
                 compression_mode=None, zip_base_archive=None):
        super(TsvNtTriplesYielder, self).__init__()
        self._source_file = source_file
        self._triples_count = 0
        self._error_triples = 0
        self._allow_untyped_numbers = allow_untyped_numbers
        self._line_reader = self._decide_line_reader(source_file=source_file,
                                                     raw_graph=raw_graph,
                                                     compression_mode=compression_mode,
                                                     zip_base_archive=zip_base_archive)
        # self.yield_triples = self._yield_triples_not_excluding_namespaces if namespaces_to_ignore is None\
        #     else self._yield_triples_excluding_namespaces


    def yield_triples(self):
        self._reset_count()
        for a_line in self._line_reader.read_lines():
            tokens = self._look_for_tokens(a_line.strip())
            if len(tokens) != 3:
                self._error_triples += 1
                log_msg(msg="This line caused error: " + a_line,
                             source=self._source_file)
            else:
                try:
                    yield (
                    tune_token(tokens[0]), tune_prop(tokens[1]), tune_token(tokens[2], allow_untyped_numbers=True))
                    self._triples_count += 1
                except ValueError as ve:
                    log_msg(msg=ve.message + "This line caused error: " + a_line,
                                 source=self._source_file)
                # if self._triples_count % 10000 == 0:
                #     print("Reading..." + self._triples_count)

    def _look_for_tokens(self, str_line):
        return str_line.split("\t")

    @property
    def yielded_triples(self):
        return self._triples_count

    @property
    def error_triples(self):
        return self._error_triples

    def _reset_count(self):
        self._error_triples = 0
        self._triples_count = 0

    # def yield_triples(self):
    #     self._reset_parsing()
    #     for a_line in self._line_reader.read_lines():
    #         tokens = self._look_for_tokens(a_line.strip())
    #         if len(tokens) != 3:
    #             self._error_triples += 1
    #             log_msg(msg="This line caused error: " + a_line,
    #                          source=self._source_file)
    #         else:
    #             try:
    #                 yield (tune_token(tokens[0]),
    #                        tune_prop(tokens[1]),
    #                        tune_token(tokens[2], allow_untyped_numbers=self._allow_untyped_numbers))
    #                 self._triples_count += 1
    #             except ValueError as ve:
    #                 log_msg(msg=ve.message + "This line caused error: " + a_line,
    #                              source=self._source_file)
    #             if self._triples_count % 10000 == 0:
    #                 print("Reading..." + self._triples_count)

    # def _yield_triples_excluding_namespaces(self):
    #     self._reset_parsing()
    #     for a_line in self._line_reader.read_lines():
    #         tokens = self._look_for_tokens(a_line.strip())
    #         if len(tokens) != 3:
    #             self._error_triples += 1
    #             log_msg(msg="This line caused error: " + a_line,
    #                          source=self._source_file)
    #         else:
    #             try:
    #                 candidate_triple = (tune_token(tokens[0]),
    #                                     tune_prop(tokens[1]),
    #                                     tune_token(tokens[2], allow_untyped_numbers=True))
    #                 if not check_if_property_belongs_to_namespace_list(str(candidate_triple[1]),
    #                                                                    namespaces=self._namespaces_to_ignore):
    #                     yield candidate_triple
    #
    #                 self._triples_count += 1
    #             except ValueError as ve:
    #                 log_msg(msg=ve.message + "This line caused error: " + a_line,
    #                              source=self._source_file)
    #             if self._triples_count % 10000 == 0:
    #                 print("Reading..." + self._triples_count)



import unittest
from shexer.shaper import Shaper
from test.const import BASE_FILES, default_namespaces, G1_ALL_CLASSES_NO_COMMENTS
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth
from shexer.consts import NT

_BASE_DIR = BASE_FILES + "untyped_numbers" + pth.sep

class TestInferNumericTypesForUntypedLiterals(unittest.TestCase):

    def test_untyped_int(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=_BASE_DIR + "g1_untyped_age.nt",
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        infer_numeric_types_for_untyped_literals=True,
                        input_format=NT,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_untyped_int_and_float(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=_BASE_DIR + "g1_untyped_age.nt",  # TODO CHANGE THIS AND MORE
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        infer_numeric_types_for_untyped_literals=True,
                        input_format=NT,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))


from shexer.utils.file import load_whole_file_content
from shexer.model.shape_map import ShapeMap, ShapeMapItem
from shexer.io.shape_map.node_selector.node_selector_parser import NodeSelectorParser
from shexer.io.shape_map.label.shape_map_label_parser import ShapeMapLabelParser
from shexer.utils.dict import reverse_keys_and_values

class ShapeMapParser(object):

    def __init__(self, namespaces_prefix_dict, sgraph):
        reversed_dict = reverse_keys_and_values(namespaces_prefix_dict)
        self._node_selector_parser = NodeSelectorParser(prefix_namespaces_dict=reversed_dict,
                                                        sgraph=sgraph)
        self._label_parser = ShapeMapLabelParser(prefix_namespaces_dict=reversed_dict)
        self._sgraph = sgraph

    def parse_shape_map(self, source_file=None, raw_content=None):
        self._check_input(source_file, raw_content)
        target_content = raw_content
        if source_file is not None:
            target_content = load_whole_file_content(source_file)
        return self._parse_shape_map_from_str(target_content)

    @staticmethod
    def _check_input(source_file, raw_content):
        if (source_file is None) == (raw_content is None):
            raise ValueError("Yoy must provide exactly one kind of input")

    def _parse_shape_map_from_str(self, raw_content):
        raise NotImplementedError("Implement this in derived classes")


####################################################

from shexer.io.json.json_loader import load_string_json

_KEY_NODE_SELECTOR = "nodeSelector"
_KEY_LABEL = "shapeLabel"


class JsonShapeMapParser(ShapeMapParser):
    """
    WARNING!! This is a toy parser. We are assuming many wel--formed stuff
    for the structure of the json itself and for the shape labels.
    Node selectors are well checked

    Example of expected format:
    [
  { "nodeSelector": "<http://data.example/node1>",
    "shapeLabel": "<http://schema.example/Shape2>"
    },
  { "nodeSelector": "<http://data.example/node1>",
    "shapeLabel": "<http://schema.example/Shape2>"
    }
]
    """

    def __init__(self, namespaces_prefix_dict, sgraph):
        super().__init__(namespaces_prefix_dict=namespaces_prefix_dict,
                         sgraph=sgraph)

    def _parse_shape_map_from_str(self, raw_content):
        result = ShapeMap()
        json_obj = load_string_json(raw_content)
        for a_list_elem in json_obj:
            result.add_item(ShapeMapItem(
                node_selector=self._node_selector_parser.parse_node_selector(a_list_elem[_KEY_NODE_SELECTOR]),
                shape_label=self._label_parser.parse_shape_map_label(a_list_elem[_KEY_LABEL])
            )
            )
        return result


####################################################


from shexer.io.line_reader.raw_string_line_reader import RawStringLineReader


class FixedShapeMapParser(ShapeMapParser):
    """
    WARNING!!!     This is a toy parser.

    Currently, this parser of Fixed ShapeMap syntax requires each couple selector@label to be in separate lines.
    Also, It will just assume trailing commas at the end of the line. If they are there, thats OK. If not, it will
    assume that its just because it is the last element.
    """

    def __init__(self, namespaces_prefix_dict, sgraph):
        super().__init__(namespaces_prefix_dict=namespaces_prefix_dict,
                         sgraph=sgraph)

    def _parse_shape_map_from_str(self, raw_content):
        result = ShapeMap()
        for a_line in RawStringLineReader(raw_string=raw_content).read_lines():
            a_line = a_line.strip()
            if not self._is_an_empty_line(a_line):
                result.add_item(self._parse_shape_map_item_from_line(a_line))

        return result

    def _is_an_empty_line(self, line):
        """
        It is expecting to receive a line which has already been stripped ()
        :param line:
        :return:
        """
        if len(line) == 0:
            return True
        if line[0] == "#":  # It is a comment
            return True
        return False

    def _parse_shape_map_item_from_line(self, line):
        """
        It is expecting to receive a line which has already been stripped ()
        :param line:
        :return:
        """
        line = self._remove_trailing_comma(line)
        pieces = line.split("@")
        if len(pieces) != 2:
            raise ValueError("There must be exactly a '@' char for each couple selector-label")
        return ShapeMapItem(shape_label=self._label_parser.parse_shape_map_label(pieces[1].strip()),
                            node_selector=self._node_selector_parser.parse_node_selector(pieces[0].strip()))

    @staticmethod
    def _remove_trailing_comma(line):
        if line[-1] == ",":
            return line[:-1]
        return line



from shexer.utils.shapes import build_shapes_name_for_class_uri
from shexer.utils.uri import remove_corners, unprefixize_uri_if_possible


def tune_target_classes_if_needed(list_target_classes, prefix_namespaces_dict):
    result = []
    for a_original_class in list_target_classes:
        if a_original_class.startswith("<"):
            result.append(remove_corners(a_uri=a_original_class))
        else:
            result.append(unprefixize_uri_if_possible(target_uri=a_original_class,
                                                      prefix_namespaces_dict=prefix_namespaces_dict,
                                                      include_corners=False))
    return result

def determine_original_target_nodes_if_needed(remove_empty_shapes, original_target_classes, original_shape_map, shapes_namespace):
    if not remove_empty_shapes:
        return None  # We dont need this structure if there are no shapes to remove.
    result = set()
    if original_target_classes is not None:
        for a_class in original_target_classes:
            result.add(build_shapes_name_for_class_uri(class_uri=a_class,
                                                       shapes_namespace=shapes_namespace))
    if original_shape_map is not None:
        for an_item in original_shape_map.yield_items():
            result.add(an_item.shape_label)
    return result


from shexer.consts import JSON, FIXED_SHAPE_MAP
from shexer.io.shape_map.shape_map_parser import JsonShapeMapParser, FixedShapeMapParser

def get_shape_map_parser(format, sgraph, namespaces_prefix_dict):
    if format == JSON:
        return JsonShapeMapParser(sgraph=sgraph,
                                  namespaces_prefix_dict=namespaces_prefix_dict)
    elif format == FIXED_SHAPE_MAP:
        return FixedShapeMapParser(namespaces_prefix_dict=namespaces_prefix_dict,
                                   sgraph=sgraph)
    else:
        raise ValueError("ShapeMap format not recognized:" + format)




from shexer.io.graph.yielder.big_ttl_triples_yielder import BigTtlTriplesYielder
from shexer.io.graph.yielder.multifile_base_triples_yielder import MultifileBaseTripleYielder


class MultiBigTtlTriplesYielder(MultifileBaseTripleYielder):

    def __init__(self, list_of_files, allow_untyped_numbers=False, compression_mode=None, zip_base_archive=None):
        super(MultiBigTtlTriplesYielder, self).__init__(list_of_files=list_of_files,
                                                        allow_untyped_numbers=allow_untyped_numbers,
                                                        zip_base_archive=zip_base_archive,
                                                        compression_mode=compression_mode)

    def _constructor_file_yielder(self, a_source_file, parse_namespaces=False):
        return BigTtlTriplesYielder(source_file=a_source_file,
                                    allow_untyped_numbers=self._allow_untyped_numbers,
                                    compression_mode=self._compression_mode,
                                    zip_base_archive=self._zip_base_archive)






import unittest
from shexer.shaper import Shaper
from test.const import BASE_FILES, default_namespaces
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth
from shexer.consts import TURTLE

_BASE_DIR = BASE_FILES + "exact_cardinality" + pth.sep


class TestDisableExactCardinality(unittest.TestCase):

    def test_exact_enabled(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "g6_exact_cardinality.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=True,
            disable_exact_cardinality=False)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g6_exact_enabled.shex",
                                                      str_target=str_result))

    def test_exact_disabled(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "g6_exact_cardinality.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=True,
            disable_exact_cardinality=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g6_exact_disabled.shex",
                                                      str_target=str_result))



import unittest
from shexer.shaper import Shaper
from test.const import BASE_FILES, default_namespaces
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth
from shexer.consts import TURTLE

_BASE_DIR = BASE_FILES + "namespaces_to_ignore" + pth.sep

class TestNamespacesToIgnore(unittest.TestCase):

    def test_excluding_other(self):

        namespaces = {key:value for key, value in default_namespaces().items()}

        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=_BASE_DIR + "g1_namespaces.ttl",
                        namespaces_dict=namespaces,
                        namespaces_to_ignore=["http://other.org/"],
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g1_other_namespace.shex",
                                                      str_target=str_result))

    def test_excluding_direct_other(self):
        namespaces = {key: value for key, value in default_namespaces().items()}

        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=_BASE_DIR + "g1_namespaces_indirect.ttl",
                        namespaces_dict=namespaces,
                        namespaces_to_ignore=["http://other.org/"],
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g1_other_namespace_indirect.shex",
                                                      str_target=str_result))


import unittest
from shexer.shaper import Shaper
from test.const import BASE_FILES
from test.t_utils import file_vs_str_tunned_comparison, text_contains_lines
import os.path as pth

_BASE_DIR = BASE_FILES + "wikidata_annotation" + pth.sep

class TestWikidataAnnotation(unittest.TestCase):

    def test_no_annotation(self):
        shaper = Shaper(all_classes_mode=True,
                        graph_file_input=_BASE_DIR + "wiki_example.ttl",
                        instantiation_property="http://www.wikidata.org/prop/direct/P31",
                        input_format="turtle",
                        disable_comments=True,
                        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "wiki_example_noanot.shex",
                                                      str_target=str_result))

    def test_annotation(self):
        shaper = Shaper(all_classes_mode=True,
                        graph_file_input=_BASE_DIR + "wiki_example.ttl",
                        instantiation_property="http://www.wikidata.org/prop/direct/P31",
                        input_format="turtle",
                        disable_comments=True,
                        wikidata_annotation=True
                        )
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(text_contains_lines(text=str_result,
                                            list_lines=[
                                                "rdfs:comment",
                                                "Q5  -->  human",
                                                "P31  -->  instance of",
                                                "Q215627  -->  person",
                                                "Q11689315  -->  (no label available)"
                                            ]))



from shexer.core.profiling.class_profiler import RDF_TYPE_STR
from shexer.utils.file import load_whole_file_content
from shexer.utils.uri import prefixize_uri_if_possible
from shexer.model.shape import STARTING_CHAR_FOR_SHAPE_NAME
import os

_MODEL_FILE_NAME = "model.yaml"
_PREFIX_FILE_NAME = "prefix.yaml"
_ENDPOINT_FILE_NAME = "endpoint.yaml"

_SHAPE_INDENT_LEVEL = 0
_PROPERTY_INDENT_LEVEL = 1
_CONSTRAINT_INDENT_LEVEL = 2


class RdfConfigSerializer(object):

    def __init__(self, target_directory, shapes_list, endpoint_url, namespaces_dict=None, string_return=False,
                 instantiation_property_str=RDF_TYPE_STR, shape_example_features=None, inverse_paths=False):
        self._target_directory = target_directory
        self._namespaces_dict = namespaces_dict if namespaces_dict is not None else {}
        self._shapes_list = shapes_list
        self._string_return = string_return
        self._instantiation_property_str = instantiation_property_str
        self._shape_example_features = shape_example_features
        self._inverse_paths = inverse_paths

        self._prefixes_file = self._generate_proper_path(_PREFIX_FILE_NAME)
        self._model_file = self._generate_proper_path(_MODEL_FILE_NAME)
        self._endpoint_file = self._generate_proper_path(_ENDPOINT_FILE_NAME)
        self._endpoint_url = endpoint_url

        self._subjects_dict = {}
        self._variables_used = set()
        self._shape_names_used = set()
        self._shapes_stream = None  # It will be initialized later, io_stream to write shapes

    def serialize_shapes(self):
        self._create_dir_if_necessary()
        self._serialize_prefixes()
        self._serialize_endpoint()
        self._serialize_shapes()
        return self._compose_output_files_in_str()

    def _serialize_shapes(self):
        self._shapes_stream = open(self._generate_proper_path(_MODEL_FILE_NAME), "w")
        for a_shape in self._shapes_list:
            self._serialize_shape(a_shape)
            self._shapes_stream.write("\n")
        self._shapes_stream.close()

    def _serialize_shape(self, shape):
        self._serialize_shape_header(shape)
        for a_constraint in shape.direct_statements:
            self._serialize_constraint(shape=shape,
                                       constraint=a_constraint)


    def _variable_property_name(self, st_property):
        original = self._create_var_name_for_property(st_property)  # TODO improve
        candidate = original
        counter = 1
        while candidate in self._variables_used:
            counter +=1
            candidate = original + str(counter)
        self._variables_used.add(candidate)
        return candidate

    def _create_var_name_for_property(self, st_property):
        st_property.replace("_","")
        st_property=st_property.lower()
        for i in range(len(st_property) - 1, -1, -1):
            if not st_property[i].isalnum():
                return st_property[i+1:]
        return "var_name"

    def _create_subject_name_for_shape(self, shape_uri):
        shape_uri.replace("_","")
        shape_uri.replace("-", "")
        for i in range(len(shape_uri) - 1, -1, -1):
            if not shape_uri[i].isalnum():
                return "Shape" + shape_uri[i+1:].capitalize()
        return "SubjectName"

    def _shape_subject_name(self, shape_uri):
        if shape_uri not in self._subjects_dict:
            origin = self._create_subject_name_for_shape(shape_uri)
            candidate = origin
            counter = 1
            while candidate in self._shape_names_used:
                counter += 1
                candidate = origin + str(counter)
            self._shape_names_used.add(candidate)
            self._subjects_dict[shape_uri] = candidate
        return self._subjects_dict[shape_uri]


    def _write_shape_line(self, content, indentation):
        indentation_str = ' ' * 2 * indentation
        self._shapes_stream.write(f"{indentation_str}- {content}\n")

    def _serialize_shape_header(self, shape):
        shape_example = self._shape_example_features.shape_example(shape_id=shape.class_uri)
        if not shape_example:
            shape_example = ""
        elif shape_example.startswith("http://") or shape_example.startswith("https://"):
            shape_example = f"<{shape_example}>"
            shape_example = prefixize_uri_if_possible(target_uri=shape_example,
                                                      namespaces_prefix_dict=self._namespaces_dict,
                                                      corners=True)
        self._write_shape_line(content=f"{self._shape_subject_name(shape.name[2:-1])} {shape_example}:",
                               indentation=_SHAPE_INDENT_LEVEL)

    def _serialize_constraint(self, shape, constraint):
        st_property = self._nice_uri(constraint.st_property)
        if constraint.st_property == self._instantiation_property_str:
            st_type = self._nice_uri(constraint.st_type)
            self._write_shape_line(indentation=_PROPERTY_INDENT_LEVEL,
                                   content=f"{st_property}: {st_type}")
        else:
            if self._inverse_paths:
                example_cons = self._shape_example_features.get_constraint_example(shape_id=shape.class_uri,
                                                                                   prop=constraint.st_property,
                                                                                   inverse=False)
            else:
                example_cons = self._shape_example_features.get_constraint_example(shape_id=shape.class_uri,
                                                                                   prop=constraint.st_property)
            if constraint.st_type.startswith(STARTING_CHAR_FOR_SHAPE_NAME):
                example_cons = self._shape_subject_name(constraint.st_type[2:-1])
            elif example_cons is None:
                example_cons = ""
            elif example_cons.startswith("http://") or example_cons.startswith("https://"):
                example_cons = self._nice_uri(example_cons)
            elif not example_cons.isnumeric():
                example_cons = f'"{example_cons}"'
            self._write_shape_line(indentation=_PROPERTY_INDENT_LEVEL,
                                   content=f"{st_property}:")
            self._write_shape_line(indentation=_CONSTRAINT_INDENT_LEVEL,
                                   content=f"{self._variable_property_name(constraint.st_property)}: {example_cons}")

    def _serialize_prefixes(self):
        with open(self._prefixes_file, "w") as out_stream:
            for a_namespace, a_prefix in self._namespaces_dict.items():
                out_stream.write(f"{a_prefix}: <{a_namespace}>\n")

    def _serialize_endpoint(self):
        if self._endpoint_url is not None:
            with open(self._endpoint_file, "w") as out_stream:
                out_stream.write(f"endpoint:\n  - {self._endpoint_url}")

    def _generate_proper_path(self, file_name):
        return self._target_directory + (
            "" if self._target_directory.endswith(os.path.sep) else os.path.sep) + file_name

    def _compose_output_files_in_str(self):
        return f"{self._header(_ENDPOINT_FILE_NAME)}\n" \
               f"{load_whole_file_content(self._endpoint_file) if self._endpoint_url is not None else 'Endpoint not provided'}\n" \
               f"{self._header(_PREFIX_FILE_NAME)}\n" \
               f"{load_whole_file_content(self._prefixes_file)}\n" \
               f"{self._header(_MODEL_FILE_NAME)}\n" \
               f"{load_whole_file_content(self._model_file)}"

    def _header(self, file_name):
        return f"##################  {file_name}  ##################"

    def _create_dir_if_necessary(self):
        if not os.path.exists(self._target_directory):
            os.makedirs(self._target_directory)

    def _nice_uri(self, target_uri):
        result = f"<{target_uri}>"
        return prefixize_uri_if_possible(target_uri=result,
                                         namespaces_prefix_dict=self._namespaces_dict,
                                         corners=True)




import unittest
from shexer.shaper import Shaper
from test.const import BASE_FILES
from test.t_utils import file_vs_str_tunned_comparison
from shexer.consts import TURTLE
import os.path as pth

_BASE_DIR = BASE_FILES + "qualifiers" + pth.sep

class TestShapeQualifiersMode(unittest.TestCase):

    def test_wikidata_virus_and_qualifiers(self):
        shape_map_raw = "SPARQL'SELECT DISTINCT ?virus WHERE {   VALUES ?virus {  wd:Q82069695  }  }'@<Virus>  "
        # wd:Q8351095  wd:Q16983356 wd:Q4902157  wd:Q278567 wd:Q16983360 wd:Q16991954

        namespaces_dict = {
            "http://www.w3.org/2000/01/rdf-schema#": "rdfs",
            "http://www.wikidata.org/prop/": "p",
            "http://www.wikidata.org/prop/direct/": "wdt",
            "http://www.wikidata.org/entity/": "wd",
            "http://www.w3.org/2001/XMLSchema#": "xsd",
            "http://www.w3.org/1999/02/22-rdf-syntax-ns#": "rdf",
            "http://www.w3.org/XML/1998/namespace": "xml",
            "http://wikiba.se/ontology#": "wikibase",
            "http://schema.org/": "schema",
            "http://www.w3.org/2004/02/skos/core#": "skos"
        }

        shaper = Shaper(shape_map_raw=shape_map_raw,
                        graph_file_input=_BASE_DIR + "virus_wikidata_depth2.ttl",
                        input_format=TURTLE,
                        all_instances_are_compliant_mode=True,
                        infer_numeric_types_for_untyped_literals=True,
                        discard_useless_constraints_with_positive_closure=True,
                        keep_less_specific=True,
                        instantiation_property="http://www.wikidata.org/prop/direct/P31",
                        namespaces_dict=namespaces_dict,
                        namespaces_for_qualifier_props=["http://www.wikidata.org/prop/"],
                        shape_qualifiers_mode=True,
                        disable_comments=True,
                        all_classes_mode=False,
                        allow_opt_cardinality=False)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "virus_qualifiers.shex",
                                                      str_target=str_result))



from shexer.utils.shapes import build_shapes_name_for_class_uri
from shexer.core.profiling.consts import POS_CLASSES, _S, _P, _O, POS_FEATURES_DIRECT, _ONE_TO_MANY, POS_FEATURES_INVERSE
from shexer.model.IRI import IRI_ELEM_TYPE, IRI
from shexer.model.bnode import BNode, BNODE_ELEM_TYPE

class AbstractFeatureDirectionStrategy(object):

    def __init__(self, class_profiler):
        self._class_profiler = class_profiler
        self._i_dict = self._class_profiler._instances_dict
        self._c_shapes_dict = self._class_profiler._classes_shape_dict
        self._c_counts = self._class_profiler._class_counts
        self._shape_names_dict = self._class_profiler._shape_names_dict
        self._original_raw_target_classes = self._class_profiler._original_raw_target_classes
        self._detect_minimal_iri = self._class_profiler._detect_minimal_iri
        self._examples_mode = self._class_profiler._examples_mode
        if self._detect_minimal_iri or self._examples_mode is not None:
            self._shape_feature_examples = self._class_profiler._shape_feature_examples

    def adapt_instances_dict(self):
        raise NotImplementedError()

    def is_a_relevant_triple(self, a_triple):
        raise NotImplementedError()

    def annotate_triple_features(self, a_triple):
        raise NotImplementedError()

    def annotate_instance_features(self, an_instance):
        raise NotImplementedError()

    def init_original_targets(self):
        raise NotImplementedError()

    def init_annotated_targets(self):
        raise NotImplementedError()

    def has_shape_annotated_features(self, shape_label):
        raise NotImplementedError()
    #
    # def look_for_example_features(self, instance_id, shape_id):
    #     raise NotImplementedError()

    def _init_annotated_direct_features(self):
        for an_instance, class_list in self._i_dict.items():
            for a_class in class_list:
                if a_class not in self._c_shapes_dict:
                    self._c_shapes_dict[a_class] = {}
                    self._c_counts[a_class] = 0
                self._c_counts[a_class] += 1

    def _annotate_direct_instance_features(self, an_instance):
        direct_feautres_3tuple = self._infer_direct_3tuple_features(an_instance)

        for a_class in self._i_dict[an_instance][POS_CLASSES]:
            self._annotate_direct_instance_features_for_class(a_class, direct_feautres_3tuple)

    def _infer_direct_3tuple_features(self, an_instance):
        result = []
        for a_prop in self._i_dict[an_instance][POS_FEATURES_DIRECT]:
            for a_type in self._i_dict[an_instance][POS_FEATURES_DIRECT][a_prop]:
                for a_valid_cardinality in self._infer_valid_cardinalities(a_prop,
                                                                           self._i_dict[an_instance][POS_FEATURES_DIRECT][a_prop][a_type]):
                    result.append( (a_prop, a_type, a_valid_cardinality) )
        return result


    def _infer_valid_cardinalities(self, a_property, a_cardinality):
        """
        Special teratment for self._instantiation_property_str. If thats the property, we are targetting specific URIs
        instead of the type IRI.
        Cardinality will be always "1"
        :param a_property:
        :param a_cardinality:
        :return:
        """
        if a_property == self._class_profiler._instantiation_property_str:
            yield 1
        else:
            yield a_cardinality
            yield _ONE_TO_MANY

    def _annotate_direct_instance_features_for_class(self, a_class, features_3tuple):
        for a_feature_3tuple in features_3tuple:
            self._introduce_needed_elements_in_shape_classes_dict(a_class, a_feature_3tuple)
            # 3tuple: 0->str_prop, 1->str_type, 2->cardinality
            self._c_shapes_dict[a_class][a_feature_3tuple[0]][a_feature_3tuple[1]][a_feature_3tuple[2]] += 1

    def _introduce_needed_elements_in_shape_classes_dict(self, a_class, a_feature_3tuple):
        str_prop = a_feature_3tuple[0]
        str_type = a_feature_3tuple[1]
        cardinality = a_feature_3tuple[2]
        if str_prop not in self._c_shapes_dict[a_class]:
            self._c_shapes_dict[a_class][str_prop] = {}
        if str_type not in self._c_shapes_dict[a_class][str_prop]:
            self._c_shapes_dict[a_class][str_prop][str_type] = {}
        if cardinality not in self._c_shapes_dict[a_class][str_prop][str_type]:
            self._c_shapes_dict[a_class][str_prop][str_type][cardinality] = 0

    def _is_relevant_instance(self, an_instance):
        return (isinstance(an_instance, IRI) or isinstance(an_instance, BNode)) and an_instance.iri in self._i_dict

    def _decide_type_elem(self, original_elem, str_prop):
        """
        Special treatment for self._instantiation_property_str property. We look for ValueSets instead of types when this property appears.

        :param original_elem:
        :param str_prop:
        :return:
        """
        if str_prop != self._class_profiler._instantiation_property_str:
            return original_elem.elem_type
        return original_elem.iri

    def _decide_shapes_elem(self, str_elem):
        if str_elem not in self._i_dict:
            return []
        return [self._get_shape_name_for_a_class(a_class)
                for a_class in self._i_dict[str_elem][POS_CLASSES]]

    def _get_shape_name_for_a_class(self, a_class):
        self._assign_shape_name_if_needed(a_class)
        return self._shape_names_dict[a_class]

    def _assign_shape_name_if_needed(self, a_class):
        if a_class in self._shape_names_dict:
            return
        self._shape_names_dict[a_class] = \
            build_shapes_name_for_class_uri(class_uri=a_class,
                                            shapes_namespace=self._class_profiler._shapes_namespace)

    def _annotate_target_subject(self, a_triple):
        str_subj = a_triple[_S].iri
        str_prop = a_triple[_P].iri
        type_obj = self._decide_type_elem(a_triple[_O], str_prop)

        obj_shapes = [] if type_obj not in [IRI_ELEM_TYPE, BNODE_ELEM_TYPE] else self._decide_shapes_elem(a_triple[_O].iri)

        self._introduce_needed_elements_in_shape_instances_dict_for_subj(str_subj=str_subj,
                                                                         str_prop=str_prop,
                                                                         type_obj=type_obj,
                                                                         obj_shapes=obj_shapes)
        self._i_dict[str_subj][POS_FEATURES_DIRECT][str_prop][type_obj] += 1
        for a_shape in obj_shapes:
            self._i_dict[str_subj][POS_FEATURES_DIRECT][str_prop][a_shape] += 1


    def _introduce_needed_elements_in_shape_instances_dict_for_subj(self, str_subj, str_prop, type_obj, obj_shapes):
        if str_prop not in self._i_dict[str_subj][POS_FEATURES_DIRECT]:
            self._i_dict[str_subj][POS_FEATURES_DIRECT][str_prop] = {}
        if type_obj not in self._i_dict[str_subj][POS_FEATURES_DIRECT][str_prop]:
            self._i_dict[str_subj][POS_FEATURES_DIRECT][str_prop][type_obj] = 0
        for a_shape in obj_shapes:
            if a_shape not in self._i_dict[str_subj][POS_FEATURES_DIRECT][str_prop]:
                self._i_dict[str_subj][POS_FEATURES_DIRECT][str_prop][a_shape] = 0


import unittest
from shexer.shaper import Shaper
from test.const import G1, BASE_FILES, G1_ALL_CLASSES_NO_COMMENTS, default_namespaces
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth
from shexer.consts import TURTLE

_BASE_DIR = BASE_FILES + "threshold" + pth.sep

class TestGraphFileInput(unittest.TestCase):

    def test_t_0(self):
        shaper = Shaper(graph_file_input=G1,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=True,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True,
                                       acceptance_threshold=0)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))
    def test_t_1(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True,
                                       acceptance_threshold=1)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g1_t1.shex",
                                                      str_target=str_result))

    def test_t_05(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True,
                                       acceptance_threshold=.5)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g1_t05.shex",
                                                      str_target=str_result))

    def test_t_051(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True,
                                       acceptance_threshold=.51)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g1_t051.shex",
                                                      str_target=str_result))



from shexer.core.profiling.class_profiler import RDF_TYPE_STR
from shexer.model.shape import STARTING_CHAR_FOR_SHAPE_NAME
from rdflib import Graph, Namespace, URIRef, RDF, BNode, XSD, Literal
from shexer.model.statement import POSITIVE_CLOSURE, KLEENE_CLOSURE, OPT_CARDINALITY
from shexer.utils.uri import XSD_NAMESPACE, LANG_STRING_TYPE
from shexer.model.const_elem_types import IRI_ELEM_TYPE, LITERAL_ELEM_TYPE, DOT_ELEM_TYPE, BNODE_ELEM_TYPE
from shexer.io.wikidata import wikidata_annotation
from wlighter import TURTLE_FORMAT
from shexer.model.node_selector import NodeSelectorSparql, NodeSelectorNoSparql
from shexer.utils.log import log_msg

_EXPECTED_SHAPE_BEGINING = STARTING_CHAR_FOR_SHAPE_NAME + "<"
_EXPECTED_SHAPE_ENDING = ">"

_SHACL_NAMESPACE = "http://www.w3.org/ns/shacl#"

_SHACL_PRIORITY_PREFIXES = ["sh", "shacl", "sha"]

_R_SHACL_SHAPE_URI = URIRef(_SHACL_NAMESPACE + "NodeShape")
_R_SHACL_PROPERTY_SHAPE_URI = URIRef(_SHACL_NAMESPACE + "PropertyShape")

_R_SHACL_TARGET_CLASS_PROP = URIRef(_SHACL_NAMESPACE + "targetClass")
_R_SHACL_TARGET_NODE_PROP = URIRef(_SHACL_NAMESPACE + "targetNode")
_R_SHACL_PATH_PROP = URIRef(_SHACL_NAMESPACE + "path")
_R_SHACL_INVERSE_PATH_PROP = URIRef(_SHACL_NAMESPACE + "inversePath")
_R_SHACL_MIN_COUNT_PROP = URIRef(_SHACL_NAMESPACE + "minCount")
_R_SHACL_MAX_COUNT_PROP = URIRef(_SHACL_NAMESPACE + "maxCount")

_R_SHACL_PROPERTY_PROP = URIRef(_SHACL_NAMESPACE + "property")

_R_SHACL_DATATYPE_PROP = URIRef(_SHACL_NAMESPACE + "dataType")
_R_SHACL_NODEKIND_PROP = URIRef(_SHACL_NAMESPACE + "nodeKind")
_R_SHACL_NODE_PROP = URIRef(_SHACL_NAMESPACE + "node")

_R_SHACL_IN_PROP = URIRef(_SHACL_NAMESPACE + "in")

_R_SHACL_PATTERN_PROP = URIRef(_SHACL_NAMESPACE + "pattern")

_R_SHACL_NODEKIND_IRI = URIRef(_SHACL_NAMESPACE + "IRI")
_R_SHACL_NODEKIND_LITERAL = URIRef(_SHACL_NAMESPACE + "Literal")
_R_SHACL_NODEKIND_BNODE = URIRef(_SHACL_NAMESPACE + "BlankNode")
_R_SHACL_NODEKIND_DOT = None

_R_LANG_STRING = URIRef("http://www.w3.org/2000/01/rdf-schema#langString")

_INTEGER = "i"
_STRING = "s"

_MACRO_MAPPING = {IRI_ELEM_TYPE: _R_SHACL_NODEKIND_IRI,
                  LITERAL_ELEM_TYPE: _R_SHACL_NODEKIND_LITERAL,
                  DOT_ELEM_TYPE: _R_SHACL_NODEKIND_BNODE,
                  BNODE_ELEM_TYPE: _R_SHACL_NODEKIND_DOT}


class ShaclSerializer(object):

    def __init__(self, target_file, shapes_list, namespaces_dict=None, string_return=False,
                 instantiation_property_str=RDF_TYPE_STR, wikidata_annotation=False,
                 detect_minimal_iri=False, shape_example_features=None, shape_map=None, verbose=False):
        self._target_file = target_file
        self._namespaces_dict = namespaces_dict if namespaces_dict is not None else {}
        self._shapes_list = shapes_list
        self._string_return = string_return
        self._instantiation_property_str = instantiation_property_str
        self._wikidata_annotation = wikidata_annotation
        self._detect_minimal_iri = detect_minimal_iri
        self._shape_example_features = shape_example_features
        self._shape_map = shape_map
        self._verbose = verbose

        self._g_shapes = Graph()

        # self._uri_dict = {}

    def serialize_shapes(self):
        self._add_namespaces()
        self._add_shapes()
        return self._produce_output()

    #################### NAMESPACES

    def _add_namespaces(self):
        self._add_param_namespaces()
        self._add_shacl_namespace_if_needed()

    def _add_param_namespaces(self):
        for a_namespace, a_prefix in self._namespaces_dict.items():
            self._add_namespace(prefix=a_prefix,
                                namespace_str=a_namespace)

    def _add_namespace(self, prefix, namespace_str):
        self._g_shapes.bind(prefix=prefix,
                            namespace=Namespace(namespace_str))

    def _add_shacl_namespace_if_needed(self):
        if _SHACL_NAMESPACE in self._namespaces_dict:  # shacl already included
            return
        curr_prefixes = self._namespaces_dict.values()
        for a_prefix in _SHACL_PRIORITY_PREFIXES:  # trying default prefixes
            if a_prefix not in curr_prefixes:
                self._add_shacl_namespace(a_prefix)
                return
        counter = 1  # going for random prefixes, no defs. available
        candidate_pref = _SHACL_PRIORITY_PREFIXES[0] + str(counter)
        while candidate_pref in curr_prefixes:
            counter += 1
            candidate_pref = _SHACL_PRIORITY_PREFIXES[0] + str(counter)
        self._add_shacl_namespace(candidate_pref)

    def _add_shacl_namespace(self, shacl_prefix):
        self._add_namespace(prefix=shacl_prefix,
                            namespace_str=_SHACL_NAMESPACE)
        self._namespaces_dict[_SHACL_NAMESPACE] = shacl_prefix

    #################### SHAPES

    def _add_shapes(self):
        for a_shape in self._shapes_list:
            self._add_shape(a_shape)

    def _add_shape(self, shape):
        r_shape_uri = self._generate_shape_uri(shape_name=shape.name)
        self._add_shape_uri(r_shape_uri=r_shape_uri)
        self._add_target_class(r_shape_uri=r_shape_uri,
                               shape=shape)
        if self._detect_minimal_iri:
            self._add_min_iri(r_shape_uri=r_shape_uri,
                              shape=shape)
        self._add_shape_constraints(shape=shape,
                                    r_shape_uri=r_shape_uri)


    def _add_target_class(self, shape, r_shape_uri):
        if self._shape_map is None:
            if shape.class_uri is not None:
                self._add_triple(r_shape_uri,
                                 _R_SHACL_TARGET_CLASS_PROP,
                                 URIRef(shape.class_uri))
        else:
            self._add_target_declaration_from_shape_map(shape, r_shape_uri)

    def _add_min_iri (self, shape, r_shape_uri):
        # if shape.iri_pattern is not None:
        if self._shape_example_features.shape_min_iri(shape_id=shape.class_uri) is not None:
            self._add_triple(r_shape_uri,
                             _R_SHACL_PATTERN_PROP,
                             self._literal_iri_pattern(shape))

    def _literal_iri_pattern(self, shape):
        return Literal("^{}".format(self._shape_example_features.shape_min_iri(shape_id=shape.class_uri)))

    def _add_shape_constraints(self, shape, r_shape_uri):
        for a_statement in shape.yield_statements():
            self._add_constraint(statement=a_statement,
                                 r_shape_uri=r_shape_uri)

    def _is_instantiation_property(self, str_property):
        return str_property == self._instantiation_property_str

    def _add_constraint(self, statement, r_shape_uri):
        if self._is_instantiation_property(statement.st_property):
            self._add_instantiation_constraint(statement=statement,
                                               r_shape_uri=r_shape_uri)
        else:
            self._add_regular_constraint(statement=statement,
                                         r_shape_uri=r_shape_uri)

    def _add_exactly_one_cardinality(self, r_constraint_node):
        self._add_min_occurs(r_constraint_node=r_constraint_node,
                             min_occurs=1)
        self._add_max_occurs(r_constraint_node=r_constraint_node,
                             max_occurs=1)

    def _add_in_instance(self, r_constraint_node, statement):
        target_node = self._generate_r_uri_for_str_uri(statement.st_type)
        list_seed_node = self._generate_bnode()
        self._add_triple(r_constraint_node, _R_SHACL_IN_PROP, list_seed_node)
        self._add_triple(list_seed_node, RDF.first, target_node)
        self._add_triple(list_seed_node, RDF.rest, RDF.nil)

    def _add_instantiation_constraint(self, statement, r_shape_uri):
        r_constraint_node = self._generate_bnode()
        self._add_bnode_property(r_shape_uri=r_shape_uri,
                                 r_constraint_node=r_constraint_node)
        self._add_direct_path(statement=statement,
                              r_constraint_node=r_constraint_node)
        self._add_exactly_one_cardinality(r_constraint_node=r_constraint_node)
        self._add_in_instance(statement=statement,
                              r_constraint_node=r_constraint_node)

    def _add_regular_constraint(self, statement, r_shape_uri):
        r_constraint_node = self._generate_bnode()
        self._add_bnode_property(r_shape_uri=r_shape_uri,
                                 r_constraint_node=r_constraint_node)
        self._add_node_type(statement=statement,
                            r_constraint_node=r_constraint_node)
        self._add_cardinality(statement=statement,
                              r_constraint_node=r_constraint_node)
        self._add_path(statement=statement,
                       r_constraint_node=r_constraint_node)

    def _add_path(self, statement, r_constraint_node):
        if not statement.is_inverse:
            self._add_direct_path(statement=statement,
                                  r_constraint_node=r_constraint_node)
        else:
            self._add_inverse_path(statement=statement,
                                   r_constraint_node=r_constraint_node)

    def _add_direct_path(self, statement, r_constraint_node):
        r_property_uri = self._generate_r_uri_for_str_uri(statement.st_property)
        self._add_triple(r_constraint_node, _R_SHACL_PATH_PROP, r_property_uri)

    def _add_inverse_path(self, statement, r_constraint_node):
        r_property_uri = self._generate_r_uri_for_str_uri(statement.st_property)
        inverse_path_node = self._generate_bnode()
        self._add_triple(r_constraint_node, _R_SHACL_PROPERTY_PROP, inverse_path_node)
        self._add_triple(inverse_path_node, _R_SHACL_INVERSE_PATH_PROP, r_property_uri)

    def _generate_r_uri_for_str_uri(self, property_str):
        if property_str.startswith("<") and property_str.endswith(">"):
            return URIRef(property_str[1:-1])
        elif property_str.startswith("http://") or property_str.startswith("https://"):
            return URIRef(property_str)
        raise ValueError("Having troubles recognizing this URI", property_str, ". "
                        "Is it well-formed? If you think so, add a GitHub issue. ")

    def _is_a_shape(self, target_type):
        return target_type.startswith(STARTING_CHAR_FOR_SHAPE_NAME)

    def _is_literal(self, target_type):
        return target_type == LANG_STRING_TYPE or target_type.startswith(XSD_NAMESPACE)

    def _is_macro(self, target_type):
        return target_type in _MACRO_MAPPING

    def _add_dataType_literal(self, r_constraint_node, target_type):
        # if target_type == LANG_STRING_TYPE:
        #     type_node = _R_LANG_STRING
        # elif target_type.endswith("integer"):
        #     type_node = XSD.integer
        # elif target_type.endswith("float"):
        #     type_node = XSD.float
        # elif target_type.endswith("string"):
        #     type_node = XSD.string
        # else:
        #     raise ValueError("Unexpected literal type:" + target_type)
        self._add_triple(r_constraint_node,
                         _R_SHACL_DATATYPE_PROP,
                         URIRef(target_type))

    def _add_node_shape(self, r_constraint_node, target_type):
        self._add_triple(r_constraint_node,
                         _R_SHACL_NODE_PROP,
                         self._generate_shape_uri(shape_name=target_type))

    def _add_nodeKind_macro(self, r_constraint_node, target_type):
        type_node = _MACRO_MAPPING[target_type]
        if type_node is not None:
            self._add_triple(r_constraint_node,
                             _R_SHACL_NODEKIND_PROP,
                             type_node)

    def _add_node_type(self, statement, r_constraint_node):
        #  sh:dataType for literal types
        #  sh:nodeKind for IRI or similar macros.
        #  sh:node for a shape
        # if self._is_literal(statement.st_type):
        #     self._add_dataType_literal(r_constraint_node=r_constraint_node,
        #                                target_type=statement.st_type)
        if self._is_macro(statement.st_type):
            self._add_nodeKind_macro(r_constraint_node=r_constraint_node,
                                     target_type=statement.st_type)
        elif self._is_a_shape(statement.st_type):
            self._add_node_shape(r_constraint_node=r_constraint_node,
                                 target_type=statement.st_type)
        else:  # It should be a literal
            self._add_dataType_literal(r_constraint_node=r_constraint_node,
                                       target_type=statement.st_type)
        # else:
        #     raise ValueError("Check here: ")


    def _min_occurs_from_cardinality(self, cardinality):
        if cardinality in [KLEENE_CLOSURE, OPT_CARDINALITY]:
            return None
        elif cardinality == POSITIVE_CLOSURE:
            return 1
        else:
            return cardinality

    def _max_occurs_from_cardinality(self, cardinality):
        if cardinality in [KLEENE_CLOSURE, POSITIVE_CLOSURE]:
            return None
        elif cardinality == OPT_CARDINALITY:
            return 1
        else:
            return cardinality

    def _generate_r_literal(self, value, l_type):
        return Literal(value, datatype=self._map_rdflib_datatype(l_type))

    def _map_rdflib_datatype(self, l_type):
        if l_type == _INTEGER:
            return XSD.integer
        elif l_type == _STRING:
            return XSD.string
        else:
            raise ValueError("Having troubles recognizing this literal type", l_type, ". "
                        "Is it well-formed? If you think so, add a GitHub issue. ")

    def _add_min_occurs(self, r_constraint_node, min_occurs):
        self._add_triple(r_constraint_node,
                         _R_SHACL_MIN_COUNT_PROP,
                         self._generate_r_literal(value=min_occurs,
                                                  l_type=_INTEGER))

    def _add_max_occurs(self, r_constraint_node, max_occurs):
        self._add_triple(r_constraint_node,
                         _R_SHACL_MAX_COUNT_PROP,
                         self._generate_r_literal(value=max_occurs,
                                                  l_type=_INTEGER))

    def _add_cardinality(self, statement, r_constraint_node):
        min_occurs = self._min_occurs_from_cardinality(statement.cardinality)
        max_occurs = self._max_occurs_from_cardinality(statement.cardinality)
        if min_occurs is not None:
            self._add_min_occurs(r_constraint_node=r_constraint_node,
                                 min_occurs=min_occurs)
        if max_occurs is not None:
            self._add_max_occurs(r_constraint_node=r_constraint_node,
                                 max_occurs=max_occurs)

    def _add_bnode_property(self, r_shape_uri, r_constraint_node):
        self._add_triple(r_shape_uri, _R_SHACL_PROPERTY_PROP, r_constraint_node)
        self._add_triple(r_constraint_node, RDF.type, _R_SHACL_PROPERTY_SHAPE_URI)

    def _generate_shape_uri(self, shape_name):
        if shape_name.startswith(_EXPECTED_SHAPE_BEGINING) and shape_name.endswith(_EXPECTED_SHAPE_ENDING):
            return URIRef(shape_name[2:-1])  # Excluding  "@<"  and ">
        raise ValueError("Unknown error, having trouble with a shape label:", shape_name,
                         "Add a GitHub issue to github with your input to have this review and fixed.")

    def _add_shape_uri(self, r_shape_uri):
        self._add_triple(r_shape_uri, RDF.type, _R_SHACL_SHAPE_URI)

    def _add_triple(self, s, p, o):
        self._g_shapes.add((s, p, o))

    @staticmethod
    def _generate_bnode():
        return BNode()

    #################### OUTPUT

    def _produce_output(self):
        if self._wikidata_annotation:
            return self._produce_wikidata_annotation_output()
        # destination = None if self._string_return else self._target_file
        if self._string_return:
            return self._g_shapes.serialize(format="turtle")
        else:
            self._g_shapes.serialize(destination=self._target_file, format="turtle")


    def _produce_wikidata_annotation_output(self):
        result = self._g_shapes.serialize(format="turtle")
        result = wikidata_annotation(raw_input=result,
                                     string_return=self._string_return,
                                     out_file=self._target_file,
                                     format=TURTLE_FORMAT,
                                     rdfs_comments=False)
        if self._string_return:
            return result


    ########################## shape map HACKS

    def _locate_shape_map_selector_in_shape_map(self, shape):
        for an_item in self._shape_map.yield_items():
            if shape.class_uri == an_item.shape_label:
                return an_item.node_selector

    def _add_target_declaration_from_shape_map(self, shape, r_shape_uri):
        shape_map_selector = self._locate_shape_map_selector_in_shape_map(shape)
        if type(shape_map_selector) == NodeSelectorSparql or shape_map_selector.raw_selector.startswith("{"):
            log_msg(msg=f"WARNING: You are serializing to SHACL a shabe with label {shape.class_uri} whose target instances "
                    f"are expressed with an SPARQL selector or a FOCUS expressions in a ShEx's shape map. As these selectors"
                    f"are not supported in SHACL core, the target expression will consist of the list of target nodes"
                    f"it is obtained when executing the shape maps selector's expression.", err=True, verbose=self._verbose)
        for a_node in shape_map_selector.get_target_nodes():
            self._add_triple(r_shape_uri,
                             _R_SHACL_TARGET_NODE_PROP,
                             URIRef(a_node))


import unittest
from shexer.shaper import Shaper
from test.const import G1, BASE_FILES, default_namespaces
from test.t_utils import file_vs_str_tunned_comparison, graph_comparison_file_vs_str
import os.path as pth

from shexer.consts import TURTLE, SHACL_TURTLE



_BASE_DIR = BASE_FILES + "min_iri" + pth.sep  # We just need something with another instantiation property


class TestDetectMinimalIri(unittest.TestCase):

    def test_all_classes_g1_enabled(self):
        shaper = Shaper(
            graph_file_input=G1,
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=True,
            detect_minimal_iri=True)
        str_result = shaper.shex_graph(string_output=True,
                                       output_format=SHACL_TURTLE)
        self.assertTrue(graph_comparison_file_vs_str(file_path=_BASE_DIR + "shacl_g1_all_classes_no_comments_min_iri.ttl",
                                                     str_target=str_result))

    def test_g1_different_namespaces_per_class(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR+"g1_different_namespaces_per_class.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=True,
            detect_minimal_iri=True)
        str_result = shaper.shex_graph(string_output=True,
                                       output_format=SHACL_TURTLE)
        self.assertTrue(graph_comparison_file_vs_str(file_path=_BASE_DIR + "shacl_g1_different_namespaces_per_class.ttl",
                                                     str_target=str_result))


    def test_g1_different_namespaces_per_instance(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR+"g1_different_namespaces_per_instance.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=True,
            detect_minimal_iri=True)
        str_result = shaper.shex_graph(string_output=True,
                                       output_format=SHACL_TURTLE)
        self.assertTrue(graph_comparison_file_vs_str(file_path=_BASE_DIR + "shacl_g1_different_namespaces_per_instance.ttl",
                                                     str_target=str_result))

    def test_g1_different_base_per_instance_no_sep_char(self):
        """
        :Person should be [http://example.org/ns1/], not [http://example.org/ns1/aa]
        :return:
        """
        shaper = Shaper(
            graph_file_input=_BASE_DIR+"g1_different_base_per_instance_no_sep_char.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=True,
            detect_minimal_iri=True)
        str_result = shaper.shex_graph(string_output=True,
                                       output_format=SHACL_TURTLE)
        self.assertTrue(graph_comparison_file_vs_str(file_path=_BASE_DIR + "shacl_g1_different_namespaces_per_class.ttl",
                                                     str_target=str_result))

